""" 
One symbol lookahead stream.  A lookahead does not automatically move forward 
when a symbol is read, thus allowing multiple access of that symbol.
This is extremely convenient for LL(1) parsers.  The current symbol of a 
lookahead is always defined, but it may be the END_OF_STREAM symbol.
"""

# author R.N.Bosworth

# version 8 Mar 23  15:51

from guibits1_0 import type_checking2_0, unicoding3_0
from .import unicode_io

"""
Copyright (C) 2012,2015,2016,2017,2018,2021,2022,2023  R.N.Bosworth

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License (gpl.txt) for more details.
"""

# exported constants
# ------------------
END_OF_STREAM =  - 1

# exported types
# --------------

class Lookahead:
  pass
  

# exported procedures
# -------------------

def advance(l):
  """ 
  pre:
    l = Lookahead to be advanced
    
  post:
    l has been advanced by one symbol

  test:
    once thru
  """
  type_checking2_0.check_derivative(l,Lookahead)
  l._current_symbol = unicode_io.read(l._current_stream)
  

def current_symbol_of(l):
  """ 
  pre:
    l = Lookahead whose current symbol is to be accessed
    
  post:
    current symbol of l has been returned

  test:
    once thru
  """
  type_checking2_0.check_derivative(l,Lookahead)
  return l._current_symbol
  

def lookahead_of(fn):
  """ 
  pre:
    fn = name of file to be used for the Lookahead, as str
    
  post:
    new Lookahead has been returned

  test:
    invalid file name (invalid characters)
    valid file name but file not there
    valid existent file
  """
  type_checking2_0.check_identical(fn,str)
  fl = Lookahead()
  fl._current_stream = unicode_io.new_input_reader(fn)
  fl._current_symbol = unicode_io.read(fl._current_stream)
  return fl
  

def lookahead_of_string(s):
  """ 
  pre:
    s = string to be used for the lookahead, as unicoding3_0.String
    
  post:
    new Lookahead has been returned

  test:
    empty string
    non-empty string
  """
  type_checking2_0.check_derivative(s,unicoding3_0.String)
  fl = Lookahead()
  fl._current_stream = unicode_io.new_string_reader(s)
  fl._current_symbol = unicode_io.read(fl._current_stream)
  return fl



from enum import Enum
from . import string_parsing
from guibits1_0 import coloring
from guibits1_0 import dialoging
from guibits1_0 import font_styling
from . import html_parsing
from . import looking_ahead
from . import page_laying_out
from . import persisting
from guibits1_0 import printing
from . import rendering
from . import texting
from . import text_tokenizing
from guibits1_0 import type_checking2_0
from guibits1_0 import unicoding3_0
from guibits1_0 import windowing
from guibits1_0 import writing

# author R.N.Bosworth

# version 16 Mar 23  15:16

"""
Contractor to print texts.

Copyright (C) 2019,2021,2022,2023  R.N.Bosworth

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License (gpl.txt) for more details.
"""
""" 
  Token Grammar version 11 Feb 2019   14:43

  This grammar is in EBNF (Extended Backus-Naur Form) using the following meta-symbols:  ::==  is  (  )  |  *.  The equivalent terminal characters are called LPARENTHESIS, RPARENTHESIS, BAR and STAR.

  Non-terminals are in CamelCase.  Terminal names are in UPPER_CASE.  Other symbols represent themselves.
  
  The rules are case-sensitive.
    

  Text  ::==  Token* EndTag
  
  Token  ::==  Word | Separator | NewLine | InternalTag
  
  EndTag  ::==  < e n d >
  
  Word  ::==  NonSpaceChar NonSpaceChar*
  
  NonSpaceChar is any HtmlCharacter except LF, SPACE, TAB, FF or CR
  
  Separator  ::==  SPACE (U+0020)
  
  Newline  ::==  LF (U+000A)
  
  InternalTag  ::==   (this will be extended later)
  
  Note: Rule for HtmlCharacter is given in HTMLGrammar.
"""

# exposed procedures
# ------------------

def print_text(win,t,fn,fsize,fss):
  """ 
  pre:
    win = windowing.Window associated with text to be printed
    t = texting.Text to be printed
    fn = name of font for printing t, as str
    fsize = size of font for printing t, as float
    fss = styles of font for printing t, as font_styling.FontStyles
    
  post:
    either:
      t has been printed on a page of the current default printer,
    or:
      an exception dialog has been displayed to the user

  tests:
    empty text
    text of half a line
      END alignment
    text of two lines
      MIDDLE alignment
    text of two and a half lines
      BEGIN alignment
  """
  type_checking2_0.check_derivative(win,windowing.Window)
  type_checking2_0.check_derivative(t,texting.Text)
  type_checking2_0.check_identical(fn,str)
  type_checking2_0.check_identical(fsize,float)
  type_checking2_0.check_derivative(fss,font_styling.FontStyles)
  my_globals = persisting.get_persistents()
  # set external page dimensions
  try:
    pj = printing.set_page_dimensions(persisting.get_paper_width(my_globals),persisting.get_paper_height(my_globals),persisting.get_horizontal_indent(my_globals))
    # note: this must be changed to allow different horizontal and vertical indents
    if pj == None:
      raise Exception("Problem trying to print page.  Page was not printed.")
    
    # set up internal page
    pg = rendering.new_page(win,persisting.get_paper_width(my_globals),persisting.get_paper_height(my_globals),persisting.get_horizontal_indent(my_globals),persisting.get_horizontal_indent(my_globals),fn,fss,fsize,coloring.BLACK)
    with t._my_lock:
      # find text cursor position
      loff = texting.cursor_line_offset(t)
      cpoff = texting.cursor_code_point_offset(t)
      # set text cursor to start of text
      texting.set_cursor_start(t)
      pg._current_line = unicoding3_0.string_of("")
      pg.line_width = 0.0
      pg._current_alignment = texting.get_alignment(t)
      # scan the text token by token, printing each line
      token = text_tokenizing.get_next_token(t)
      while  not unicoding3_0.equals(token,unicoding3_0.string_of("<end>")):
        _print_token(pj,token,texting.get_alignment(t),pg)
        token = text_tokenizing.get_next_token(t)
      _print_token(pj,token,texting.get_alignment(t),pg)
      # deal with <end> token
      # end the print job
      printing.end_printing(pj)
      # restore text cursor
      texting.set_cursor(t,loff,cpoff)
      
  except Exception as ex:
    # display exception message as dialog
    dialoging.show_message_dialog(persisting.get_menu_font_size(my_globals),"Printing problem",str(ex))
      

# private procedures
# ------------------

def _de_escape(s1,s2):
  """ 
  pre:
    s1 = string which is to be appended to
    s2 = string of escaped unicode code points which is to be de-escaped and appended to s1
  post:
    s2 has been de-escaped and appended to s1
  test:
    s1 = ""
      s2 = ""
    s1 = "ab"
      s2 = "1&lt;2"
  """
  i = 0
  while i < unicoding3_0.length_of(s2):
    (code_point,i) = string_parsing.parse_html_character(s2,i)
    unicoding3_0.append(s1,code_point)


def _print_line(pj,pg):
  """ 
  pre:
    pj = current print job
    pg = page for this print
    pg._current_line = line to be printed
    pg._line_width = width of _current_line in points
    pg._text_width = width of text in points
    pg._current_alignment = alignment of line to be printed

  post:
    pg._current_line' has been printed
    pg._page_position.x_offset = 0.0
    pg._page_position.y_offset = position for next line
    pg._current_line = empty string
    pg._line_width = 0.0

  test:
    line "fred"
      _current_alignment = BEGIN
      _current_alignment = END
      _current_alignment = MIDDLE
  """
  pg._page_position.x_offset = page_laying_out.x_offset_of_line(pg._line_width,pg._text_width,pg._current_alignment)
  _try_line(pj,pg)
  pg._page_position.x_offset = 0.0
  pg._page_position.y_offset += pg._font_size
  pg._current_line = unicoding3_0.new_string()
  pg._line_width = 0.0
  

def _print_token(pj,tk,a,pg):
  """ 
  pre:
    pj = current PrintJob
    tk = token to be printed
    a = alignment of text following tk
    pg = Page on which the token is to be printed,
         with _current_line, _line_width and _current_alignment set up
  post:
    if tk is an <end> token,
      the current line has been printed
    if tk is a Separator token,
      the width in points of the current line, plus the separator, 
          has been measured, and if overflow has occurred,
            the current line has been printed
            a new empty line has been started
    if tk is a Newline token,
      the current line has been printed
      the alignment has been set to a
    if tk is a Word token,
      the width in points of the current line, 
        plus the width in points of the word,
          has been measured, and if overflow has occurred,
            any trailing space has been removed from the current line
            the current line has been printed
            a new line has been started with the word
    y-offset in page has been updated

  test:
    pg._current_alignment = BEGIN
      "the cat sat on the mat."
        "\n"
          a = MIDDLE
            "\n"
              a = END
      "the cat sat on the    \nmat.<end>"
      "the<end>"
      "the<exit>"
  """
  try:
    # first code point of token gives sort
    # Tag
    if unicoding3_0.code_point_at(tk,0) == ord('<'):
      la = looking_ahead.lookahead_of_string(tk)
      label = html_parsing.parse_tag(la)
      if unicoding3_0.equals(label,unicoding3_0.string_of("end")):
        pg._page_position.x_offset = page_laying_out.x_offset_of_line(pg._line_width,pg._text_width,pg._current_alignment)
        _try_line(pj,pg)
      else:
        raise Exception("unexpected tag label when printing text:" + unicoding3_0.python_string_of(label))
        
    # Separator
    elif unicoding3_0.code_point_at(tk,0) == ord(' '):
      space_width = writing.width_in_points_of(pg._my_window," ",pg._font_name,pg._font_styles,pg._font_size)
      #space_width = font_measuring.glyph_width_of(pg._font_metrics,ord(' '))
      if pg._line_width + space_width > pg._text_width:
        _print_line(pj,pg)
        
      else:
        unicoding3_0.append(pg._current_line,ord(' '))
        pg._line_width += space_width
        
    # Newline
    elif unicoding3_0.code_point_at(tk,0) == ord('\n'):
      _print_line(pj,pg)
      pg._current_alignment = a
      
    # Word
    else:
      pg.word_width = page_laying_out.width_in_points_of_escaped(pg._my_window,tk,pg._font_name,pg._font_styles,pg._font_size)
      if pg._line_width + pg.word_width > pg._text_width:
        last_pos = unicoding3_0.length_of(pg._current_line) - 1
        if last_pos >= 0:
          if unicoding3_0.code_point_at(pg._current_line,last_pos) == ord(' '):
            space_width = writing.width_in_points_of(pg._my_window," ",pg._font_name,pg._font_styles,pg._font_size)
            unicoding3_0.remove(pg._current_line,last_pos)
            pg._line_width -= space_width
        _print_line(pj,pg)
        _de_escape(pg._current_line,tk)
        pg._line_width = pg.word_width
      else:
        _de_escape(pg._current_line,tk)
        pg._line_width += pg.word_width
        
  except Exception as ex:
    raise Exception("Unexpected exception when printing token:"+str(ex))


def _try_line(pj,pg):
  """ 
  pre:
    pj = current print job
    pg = page to be printed
  
  post:
    the current line has been printed on the page,
      if above the bottom of the text area

  test:
    line just fits on text area
    line just overlaps end of text area
  """
  if pg._page_position.y_offset + pg._font_size <= pg._text_height:
    printing.print_string(pj,unicoding3_0.python_string_of(pg._current_line),pg._font_name,pg._font_styles,pg._font_size,pg._page_position.x_offset,pg._page_position.y_offset)



from guibits1_0 import coloring
from guibits1_0 import controlling
from guibits1_0 import cursoring
from guibits1_0 import dialoging
from enum import Enum
from guibits1_0 import file_dialoging
from guibits1_0 import font_styling
from guibits1_0 import keyboarding
from guibits1_0 import menuing
from guibits1_0 import mousing
from guibits1_0 import painting
from guibits1_0 import type_checking2_0
from guibits1_0 import unicoding3_0
from guibits1_0 import window_bounding
from guibits1_0 import windowing

from . import html_texting
from . import looking_ahead
from . import mle_emitting
from . import mle_parsing
from . import persisting
from . import rendering
from . import text_printing
from . import texting
from . import unicode_io

# author R.N.Bosworth

# version 17 Mar 23  07:04
""" 
Main contractor for Emily.  The procedure "start" runs a new instantiation of Emily.  This means that multiple instances of Emily can run on multiple virtual machines, but note that persistent variables will be common to all instances, which may lead to some interesting situations.

Copyright (C) 2015,2016,2017,2018,2019,2021,2022,2023  R.N.Bosworth

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License (gpl.txt) for more details.
"""

# constants
# ---------
VERSION = "0.9"
RELEASE_YEAR = "2023"
ABOUT_MESSAGE = "Copyright © " + RELEASE_YEAR + "  Richard Bosworth\n" + "\n" + "This program is free software: you can redistribute it and/or modify " + "it under the terms of the GNU General Public License as published by " + "the Free Software Foundation, either version 3 of the License, or " + "(at your option) any later version.\n" + "\n" + "This program is distributed in the hope that it will be useful, " + "but WITHOUT ANY WARRANTY; without even the implied warranty of " + "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the " + "GNU General Public License for more details.\n" + "\n" + "You should have received a copy of the GNU General Public License " + "along with this program.  If not, see <http://www.gnu.org/licenses/>."
DEFAULT_FONT_NAME = "Times New Roman"
DEFAULT_FONT_SIZE = 12.0
PARCHMENT = coloring.new_color(1.0,1.0,0.9)

# variables
# ---------
my_persistents = None  # cache for persistents
my_window = None  # current main window
my_text = None  # current text being edited
my_page = None  # page on which text is rendered
is_loaded = False  # True iff file is loaded
file_name = None  # current file name or None if none

curpos = rendering.new_cursor_position()
# position of start and end of cursor

pivot = rendering.new_cursor_position()
# position of start of mouse-drag

text_font_style = font_styling.new_font_styles()
current_alignment = texting.Alignment.BEGIN
"""  Invariant: current_alignment == alignment of paragraph containing cursor """


"""  Note that a glyph is anything that separates two cursor positions,
     so ' ' and '\n' are glyphs, by our definition """


# main procedure
# --------------

def start(args):
  """ 
  pre:
    args = argument list, with name of this program and possible name of file to be opened
  
  post:
    Emily session has been set up
    font name and size for text have been set up,
      either from the input file, or by default
    the persistent variables for the session have been set up
      either to their default values, or to the values from 
        "EmilyPersistent.json" in the user's default directory
    if args has name of a valid .mle file to be opened, 
      the appropriate file has been loaded as the text
    or
      a new empty text has been set up
    if the supplied filename was invalid,
      an exception has been raised
    the text has been displayed on the screen
    the cursor has been set to the end of the text
  __________________________________________________________ 
  pre:
   file "EmilyPersistent.json" may be present in the user's default directory
  
  post:
    my_window has been set up
    my_persistents has been set up, either from the input file or the defaults
    my_text has been set to a new text
    my_text has been set unmodified
    my_page has been set up from the Persistents or defaults
    the zoom factor of the main window has been set 
    file "EmilyPersistent.json" in the user's default directory has been updated
      with the latest values of the persistent variables
    is_loaded = true, iff a file has been loaded
    file_name = name of current file being edited, or None if none
    curpos = the current cursor position
    the initial screen has been displayed

  test:
    null argument list
    invalid file name
    valid file name
  """
  global file_name
  global my_persistents
  global my_text
  try:
    type_checking2_0.check_derivative(args,list)
    my_persistents = persisting.get_persistents()
    my_text = texting.new_text()
    with my_text._my_lock:
      texting.set_unmodified(my_text)
    if len(args) > 1:
      file_name = args[1]
      file_name = check_file_type(file_name)
    else:
      file_name = None
    build_u_i()
  except Exception as ex:
    dialoging.show_message_dialog(persisting.get_menu_font_size(my_persistents),"Emily: error in input file name", str(ex))
      
  
# callback procedures
# -------------------

def about_item_hit(win,x,y,l):
  """ 
  pre:
    the menu item associated with this listener has been hit by the user
    my_persistents = persistent variables for the Emily app
  
  post:
    the about-pane has been displayed on the screen

  test:
    once thru
  """
  global my_persistents
  dialoging.show_message_dialog(persisting.get_menu_font_size(my_persistents),title_of(None),ABOUT_MESSAGE)


def action_key_hit(ac):
  """ 
  pre:
    ac = action code for gesture performed by user
    my_window = window for this app
    my_text = current text
    my_page = current page
    my_persistents = persistent variables for the Emily app
    curpos = current cursor position in text and on screen

  post:
    appropriate action has been carried out
    current_alignment has been set to alignment of paragraph in which cursor resides

  test:
    BACKSPACE
      cursor in paragraph (check state of Save menu item)
      cursor at start of paragraph
      cursor at start of text
    DELETE
      cursor in paragraph (check state of Save menu item)
      cursor at end of paragraph
      cursor at end of text
    ENTER
      (check state of Save menu item)
    UP_ARROW
      cursor beyond end of line above
      cursor in middle of line above
      cursor on top line
    DOWN_ARROW
      cursor beyond end of line below
      cursor in middle of line below
      cursor on bottom line
    LEFT_ARROW
      cursor in middle of line
      cursor at start of line
      cursor at start of text
    RIGHT_ARROW
      cursor in middle of line
      cursor at end of line
      cursor at end of text
    CTRL_S
  """
  global curpos
  global current_alignment
  global my_page
  global my_text
  global my_persistents
  if ac == controlling.ActionCode.BACKSPACE:
    with my_text._my_lock:
      rendering.set_text_to_on_screen_start(curpos,my_text)
      if  not html_texting.retreat(my_text):
        print('\u0007')  # BEL
      else:
        html_texting.delete_after(my_text)
        current_alignment = texting.get_alignment(my_text)
        rendering.set_on_screen_to_text(my_text,curpos)
        # curpos has text position set
        start_line = int(curpos._start_page_position.y_offset/my_page._font_size) -1
        if start_line < 0:
          start_line = 0
        rendering.render_lines(my_text,curpos,my_page,start_line,rendering.TO_END)
        # curpos has text and page positions set
        rendering.render_thin_cursor(curpos,my_page)
        texting.set_modified(my_text)
             
  elif ac == controlling.ActionCode.DELETE:
    with my_text._my_lock:
      rendering.set_text_to_on_screen_end(curpos,my_text)
      if  not html_texting.advance(my_text):
        print('\u0007')  # BEL
      else:
        html_texting.retreat(my_text)
        # grand old Duke of York
        html_texting.delete_after(my_text)
        # alignment will always be the same
        render_modified_text()
        
  elif ac == controlling.ActionCode.ENTER:
    with my_text._my_lock:
      rendering.set_text_to_on_screen_end(curpos,my_text)
      texting.insert_code_point(my_text,ord('\n'))
      # internal hard line break
      # note:alignment will always be the same
      render_modified_text()
      
  elif ac == controlling.ActionCode.UP_ARROW:
    if rendering.is_thin(curpos):
      if not rendering.move_cursor_up(curpos,my_page):
        print('\u0007')  # BEL
    else:  # fat cursor
      painting.clear_rectangles(my_window)
      rendering.collapse_to_start(curpos)
      plo1 = rendering.start_page_line_of(curpos,my_page)
      rendering.render_lines(my_text,curpos,my_page,plo1,plo1+1)
      # render the page line of the cursor only
      
  elif ac == controlling.ActionCode.DOWN_ARROW:
    if rendering.is_thin(curpos):
      if not rendering.move_cursor_down(curpos,my_page):
        print('\u0007')  # BEL
    else:  # fat cursor
      painting.clear_rectangles(my_window)
      rendering.collapse_to_end(curpos)
      plo1 = rendering.start_page_line_of(curpos,my_page)
      rendering.render_lines(my_text,curpos,my_page,plo1,plo1+1)
      # render the page line of the cursor only

  elif ac == controlling.ActionCode.LEFT_ARROW:
    with my_text._my_lock:
      rendering.set_text_to_on_screen_start(curpos,my_text)
      if rendering.is_thin(curpos):
        if not html_texting.retreat(my_text):
          print('\u0007')  # BEL
        else:
          rendering.set_on_screen_to_text(my_text,curpos)
          plo1 = rendering.start_page_line_of_back(curpos,my_page)
          rendering.render_lines(my_text,curpos,my_page,plo1,plo1+2)
          # render the present and possible previous cursor page lines
          current_alignment = texting.get_alignment(my_text)
      else:  # fat cursor
        painting.clear_rectangles(my_window)
        rendering.set_on_screen_to_text(my_text,curpos)
        plo1 = rendering.start_page_line_of(curpos,my_page)
        rendering.render_lines(my_text,curpos,my_page,plo1,plo1+1)
        # render the page line of the cursor only
  
  elif ac == controlling.ActionCode.RIGHT_ARROW:
    with my_text._my_lock:
      rendering.set_text_to_on_screen_end(curpos,my_text)
      if rendering.is_thin(curpos):
        if  not html_texting.advance(my_text):
          print('\u0007')  # BEL
        else:
          rendering.set_on_screen_to_text(my_text,curpos)
          plo1 = rendering.start_page_line_of(curpos,my_page)
          rendering.render_lines(my_text,curpos,my_page,plo1,plo1+2)
          # render the present and possible previous cursor page lines
          current_alignment = texting.get_alignment(my_text)
      else:  # fat cursor
        painting.clear_rectangles(my_window)
        rendering.set_on_screen_to_text(my_text,curpos)
        plo1 = rendering.start_page_line_of(curpos,my_page)
        rendering.render_lines(my_text,curpos,my_page,plo1,plo1+1)
        # render the page line of the cursor only
  
  elif ac == controlling.ActionCode.CTRL_S:
    save_file(persisting.get_current_file_name(my_persistents))
      
  else:
    pass


def alignment_bar_item_hit(win,x,y,l):
  """ 
  pre:
    win = window to which this menu is to be added
    (x,y) = position of mouse hit relative to window contents
    my_persistents = persistent variables for the Emily app
    my_text = current text
    curpos = current cursor position in page
  
  post: 
    alignment menu has been built and displayed in win
    current_alignment = alignment of line containing start of cursor in my_text

  test:
    once thru
  """
  global curpos
  global current_alignment
  global my_persistents
  global my_text
  rendering.set_text_to_on_screen_start(curpos,my_text)
  current_alignment = texting.get_alignment(my_text)
  sm = menuing.new_menu(win)
  fs = persisting.get_menu_font_size(my_persistents)
  label = build_tagged_label("Begin",current_alignment == texting.Alignment.BEGIN)
  menuing.add_menu_item(sm,fs,label,begin_alignment_item_hit)
  label = build_tagged_label("Middle",current_alignment == texting.Alignment.MIDDLE)
  menuing.add_menu_item(sm,fs,label,middle_alignment_item_hit)
  label = build_tagged_label("End",current_alignment == texting.Alignment.END)
  menuing.add_menu_item(sm,fs,label,end_alignment_item_hit)
  menuing.display(sm,win,x,y)
    

def begin_alignment_item_hit(win,x,y,l):
  """ 
  pre:
    curpos = position of cursor on the page
    my_text = text being edited
    my_page = page to which my_text is rendered
  
  post: 
    current_alignment =  BEGIN
    the alignment of the paragraph containing the cursor 
      in my_text has been set to BEGIN
    the text and cursor have been re-rendered on my_page

  test:
    once thru
  """
  set_current_alignment(texting.Alignment.BEGIN)
    

def character_key_hit(cp):
  """ 
  pre:
    cp = code point of character key that has just been hit by the user
    my_text = current text
    curpos = current cursor position in text and on screen

  post:
    cp has been inserted in text at required position,
      as an HtmlCharacter
    text and cursor have been re-rendered on screen

  syntax:
    HTML Grammar version 1 Dec 2018   15:05

    This grammar defines the basic HTML constructs common to any HTML file.

    This grammar is in EBNF (Extended Backus-Naur Form) using the following meta-symbols:  ::==  is  (  )  |  *.  The equivalent terminal characters are called LPARENTHESIS, RPARENTHESIS, BAR and STAR.

    Non-terminals are in CamelCase.  Terminal names are in UPPER_CASE.  Other symbols represent themselves.
    
    The rules are case-sensitive, unlike HTML5.
    
    HtmlCharacter  ::==  ( & a m p ;
                         | & g t ;
                         | & l t ;
                         | any Unicode code point except '&' (U+26), '>' (U+3E) and '<' (U+3C)
  test:
    cp == 'a' (check Save button on menu)
    cp = '&'
    cp = '<'
    cp = '>'
    cp = '%'
  """
  global my_text
  with my_text._my_lock:
    rendering.set_text_to_on_screen_end(curpos,my_text)
    # insert HtmlCharacter
    if cp == ord('>'):
      texting.insert_code_point(my_text,ord('&'))
      texting.insert_code_point(my_text,ord('g'))
      texting.insert_code_point(my_text,ord('t'))
      texting.insert_code_point(my_text,ord(';'))
      
    elif cp == ord('<'):
      texting.insert_code_point(my_text,ord('&'))
      texting.insert_code_point(my_text,ord('l'))
      texting.insert_code_point(my_text,ord('t'))
      texting.insert_code_point(my_text,ord(';'))
      
    elif cp == ord('&'):
      texting.insert_code_point(my_text,ord('&'))
      texting.insert_code_point(my_text,ord('a'))
      texting.insert_code_point(my_text,ord('m'))
      texting.insert_code_point(my_text,ord('p'))
      texting.insert_code_point(my_text,ord(';'))
      
    else:
      texting.insert_code_point(my_text,cp)
      
    render_modified_text()
      

def end_alignment_item_hit(win,x,y,l):
  """ 
  pre:
    curpos = position of cursor on the page
    my_text = text being edited
    my_page = page to which my_text is rendered

  post: 
    current_alignment =  END
    the alignment of the paragraph containing the cursor 
      in my_text has been set to END
    the text and cursor have been re-rendered on my_page

  test:
    once thru
  """
  set_current_alignment(texting.Alignment.END)
    
  
def file_bar_item_hit(win,x,y,l):
  """ 
  pre:
    win = window to which this menu is to be added
    (x,y) = position of mouse hit relative to window contents
    my_text = current text
    is_loaded = true, iff current text was loaded from a file
    my_persistents = persistent variables for the Emily app

  post: 
    file menu has been built and displayed in win

  test:
    new text (not loaded)
    text loaded, not modified
    text loaded, modified
  """
  global my_text
  global my_persistents
  fm = menuing.new_menu(win)
  fs = persisting.get_menu_font_size(my_persistents)
  menuing.add_menu_item(fm,fs,"New",new_item_hit)
  menuing.add_menu_item(fm,fs,"Load",load_item_hit)
  with my_text._my_lock:
    text_has_been_modified = texting.has_been_modified(my_text)
  if text_has_been_modified:
    l = save_item_hit
  else:
    l = None
  menuing.add_menu_item(fm,fs,"Save (Ctrl-S)",l)
  menuing.add_menu_item(fm,fs,"Save As",save_as_item_hit)
  menuing.add_menu_item(fm,fs,"Print",print_item_hit)
  menuing.display(fm,win,x,y)
    
  
def font_bar_item_hit(win,x,y,l):
  """ 
  pre:
    win = window to which this menu is to be added
    (x,y) = position of mouse hit relative to window contents
    my_persistents = persistent variables for the Emily app

  post: 
    font menu has been built and added to win    

  test:
    once thru
  """
  global my_persistents
  fs = persisting.get_menu_font_size(my_persistents)
  fom = menuing.new_menu(win)
  add_font_menu_item(fs,"Arial",fom)
  add_font_menu_item(fs,"Courier New",fom)
  add_font_menu_item(fs,"Times New Roman",fom)
  menuing.add_separator(fom)
  MINIMUM_FONT_SIZE = 9
  MAXIMUM_FONT_SIZE = 48
  STEP1 = 12
  STEP2 = 20
  pfs = MINIMUM_FONT_SIZE
  inc = 1
  while pfs <= MAXIMUM_FONT_SIZE:
    label = build_tagged_font_size_label(pfs,int(persisting.get_text_font_size(my_persistents)))
    menuing.add_menu_item(fom,fs,label,text_font_size_item_hit)
    if pfs == STEP1 or pfs == STEP2:
      inc = inc + inc
    pfs = pfs + inc
  menuing.display(fom,win,x,y)


def font_name_item_hit(win,x,y,l):
  """
  pre:
    l = font name of menu item, possibly preceded by dots and spaces, as str
    my_text = current text
    the menu item associated with this listener has been hit by the user
    my_persistents = persistent variables for the Emily app
    my_page = current page being edited
    curpos = current cursor position

  post:
    the persistent text font name has been modified
    the text in my_window has been redisplayed

  test:
    once thru
  """
  global curpos
  global my_page
  global my_persistents
  global my_text
  fname = name_of(l)
  persisting.set_text_font_name(my_persistents,fname)
  rendering.set_text_font_name(my_page,fname)
  # re-display text and cursor
  with my_text._my_lock:
    rendering.set_text_to_on_screen_end(curpos,my_text)
    rendering.set_on_screen_to_text(my_text,curpos)
    # curpos has text position set
    windowing.clear(win)
    rendering.render_lines(my_text,curpos,my_page,0,rendering.TO_END)
    # curpos has text and page positions set
    rendering.render_thin_cursor(curpos,my_page)
    texting.set_modified(my_text)


def get_latest_version_item_hit(win,x,y,l):
  """ 
  pre:
    the menu item associated with this listener has been hit by the user
    my_persistents = persistent variables for the Emily app

  post:
    Information to get the latest version of Emily has been displayed to the user.

  test:
    once thru
  """
  global _my_persistents
  dialoging.show_message_dialog(persisting.get_menu_font_size(my_persistents),"Get latest version","To download the latest version of Emily, go to\n\n" + "https://sourceforge.net/projects/emilyeditor/?source=directory")


def load_item_hit(win,x,y,l):
  """ 
  pre:
    the menu item associated with this listener has been hit by the user
    my_text = current texting.Text
    is_loaded = True, iff current text has been loaded from a file
    my_persistents = persistent variables for the Emily app
    
  post:
    the user has been asked to save the current text, if necessary
    the user has been asked for a file to load
    if the user agrees, the file has been loaded

  test:
    no file to be saved
      user wants to load
    file to be saved
      user declines to load
  """
  global my_persistents
  global my_text
  try:
    # save current text if necessary
    check_save(my_text)
      
    # ask user which file to load
    lf = file_dialoging.show_open_file_dialog(persisting.get_menu_font_size(my_persistents),"Load file",persisting.get_current_file_name(my_persistents),file_dialoging.SortMode.LATEST_FIRST)
    if lf == None:
      raise Exception("Problem opening file. File has not been loaded.")    
    else:
      lf = check_file_type(lf)
      if lf != None:
        load_file(lf)
  except Exception as ex:
    dialoging.show_message_dialog(persisting.get_menu_font_size(my_persistents),"Emily: error while loading input file", str(ex))
    
      
def menu_font_size_menu_item_hit(win,x,y,l):
  """ 
  pre:
    the menu item associated with this listener has been hit by the user
    win = window in which Menu Font Size menu is to be displayed
    (x,y) = position in window contents of hit, as floats
    my_persistents = persistent variables for the Emily app

  post:
    the Menu Font Size menu has been displayed at (x,y) in win,
      using the font size given in the persistents file

  test:
    once thru
  """
  global _my_persistents
  fs = persisting.get_menu_font_size(my_persistents)
  mfsm = build_menu_font_size_menu(fs)
  menuing.display(mfsm,win,x,y)
    

def menu_font_size_item_hit(win,x,y,l):
  """
  pre:
    menu item associated with this callback procedure has been hit by user
    win = windowing.Window where mouse-hit occurred
    x,y = x and y offsets of mouse hit from top left-hand corner of window's contents, in points as float
    l = label of menu item hit by the user, giving required font size as str
    my_persistents = persistent variables for the Emily app
    
  post:
    menu font size has been set to value specified by user
    menu bar and zoom buttons have been modified to the new menu font size
    the font size has been saved in the persistents file

  test:
    once thru
  """
  my_font_size = float(int_of(l))
  global my_persistents
  persisting.set_menu_font_size(my_persistents,my_font_size)
  menuing.set_menu_bar_font_size(win,my_font_size)
  windowing.set_font_size(win,my_font_size)
  
  
def middle_alignment_item_hit(win,x,y,l):
  """ 
  pre:
    curpos = position of cursor on the page
    my_text = text being edited
    my_page = page to which my_text is rendered

  post: 
    current_alignment =  MIDDLE
    the alignment of the paragraph containing the cursor in my_text has been set to MIDDLE
    the text and cursor have been re-rendered on my_page

  test:
    once thru
  """
  set_current_alignment(texting.Alignment.MIDDLE)


def new_item_hit(win,x,y,l):
  """ 
  pre:
    the menu item associated with this listener has been hit by the user
    win = main window of this app
    my_text = current texting.Text
    my_page = current Page
    is_loaded = True, iff current text is loaded

  post:
    the user has been asked to save the current text, if necessary
    a blank Page has been created and displayed

  test:
    no current text
    current text 
      not loaded
        unmodified
        modified
      loaded
        not modified
        modified
  """
  global curpos
  global is_loaded
  global my_page
  global my_persistents
  global my_text
  # save current text if necessary
  check_save(my_text)
    
  # set up new text
  with my_text._my_lock:
    my_text = texting.new_text()
    texting.set_unmodified(my_text)
  is_loaded = False
  # remove any old file name from title
  windowing.set_title(win,title_of(None))
  # clear window
  windowing.clear(win)
  # set the page's text cursor position to (0,0)
  my_page._text_position.line_offset = 0
  my_page._text_position.code_point_offset = 0
  # set the page's page cursor position to (0.0,0.0)
  my_page._page_position.x_offset = 0.0
  my_page._page_position.y_offset = 0.0
  # set cursor start and end to start of text
  rendering.set_cursor_position_to_start(curpos)
  # draw cursor at origin
  cursoring.draw_cursor(win,persisting.get_text_font_size(my_persistents),persisting.get_horizontal_indent(my_persistents),persisting.get_vertical_indent(my_persistents),coloring.BLACK)
    
  
def print_item_hit(win,x,y,l):
  """ 
  pre:
      the menu item associated with this listener has been hit by the user
      win = current window
      my_text = text to be printed
      persisting.get_text_font_name(my_persistents) = font name to be used
      persisting.get_text_font_size(my_persistents) = font size to be used

  post:
      the current text has been printed on the default printer

  test:
    blank text
      Printer problem causing exception
      Printer OK
    text of two or more lines
  """
  global my_persistents
  global my_text
  fss = font_styling.new_font_styles()
  text_printing.print_text(win,my_text,persisting.get_text_font_name(my_persistents),persisting.get_text_font_size(my_persistents),fss)
    
  
def save_as_item_hit(win,x,y,l):
  """ 
  pre:
      the menu item associated with this listener has been hit by the user

  post:
      the user has been asked for the name to save the current text as
      if the user agrees, the text has been saved under the specified name
      The user has been informed of successful or unsuccessful saving

  test:
    user agrees
    user disagrees
  """
  save_file_as()
    
  
def save_item_hit(win,x,y,l):
  """ 
  pre:
    the menu item associated with this listener has been hit by the user
    persisting.get_current_file_name(my_persistents) = default file name

  post:
      the current text has been saved under its default file name

  test:
    once thru
  """
  global my_persistents
  save_file(persisting.get_current_file_name(my_persistents))
    
  
def setup_bar_item_hit(win,x,y,l):
  """ 
  pre:
    win = window to which this menu is to be added
    (x,y) = position of mouse hit relative to window contents
    my_persistents = persistent variables for the Emily app

  post: 
    setup menu has been built and displayed in win

  test:
    once thru
  """
  sm = menuing.new_menu(win)
  global my_persistents
  fs = persisting.get_menu_font_size(my_persistents)
  menuing.add_menu_item(sm,fs,"Menu Font Size",menu_font_size_menu_item_hit)
  menuing.add_menu_item(sm,fs,"Get latest version",get_latest_version_item_hit)
  emily_version = "About Emily " + VERSION
  menuing.add_menu_item(sm,fs,emily_version,about_item_hit)
  menuing.display(sm,win,x,y)
    
    
def text_font_size_item_hit(win,x,y,l):
  """ 
  pre:
    l = font size of menu item, as str
    the menu item associated with this listener has been hit by the user
    my_persistents = persistent variables for the Emily app
    my_page = current page being edited
    my_text = current text being edited
    
  post:
    the persistent text font size has been modified
    the text in my_page has been redisplayed
    curpos = current cursor position

  test:
    once thru
  """
  global curpos
  global my_page
  global my_persistents
  global my_text
  fsize = float(int_of(l))
  persisting.set_text_font_size(my_persistents,fsize)
  rendering.set_text_font_size(my_page,fsize)
  # re-display text and cursor
  with my_text._my_lock:
    rendering.set_text_to_on_screen_end(curpos,my_text)
    rendering.set_on_screen_to_text(my_text,curpos)
    # curpos has text position set
    windowing.clear(win)
    rendering.render_lines(my_text,curpos,my_page,0,rendering.TO_END)
    # curpos has text and page positions set
    rendering.render_thin_cursor(curpos,my_page)
    texting.set_modified(my_text)


def window_closing(win):
  """ 
  pre:
    win = window associated with this callback
    the user has pressed the Close button of win
    my_text = Texting.Text being edited
    my_persistents = persistent variables for the Emily app
    
  post:
    returns true iff app is to close
    if true is returned, clean-up has been done prior to shutdown 

  test:
    text to be saved
    no text to be saved
  """
  global my_persistents
  global my_text
  check_save(my_text)
  wb = persisting.get_window_bounds(my_persistents)
  persisting.set_window_bounds(my_persistents,window_bounding.get_bounds(win))
  wb2 = persisting.get_window_bounds(my_persistents)
  persisting.set_zoom_factor(my_persistents,windowing.zoom_factor_of(win))
  return True


def window_opening(win):
  """
  pre:
    win = window associated with this callback
    curpos = cursor position in Text and Page
    file_name = name of file to be loaded, or None if none
    my_persistents = persistent variables for the Emily app
    text_font_style = text font style for this file
    
  post:
    my_page has been set up according to the persistent variables
    file_name, if not None, has been loaded
    is_loaded = True, iff file has been loaded
    the text, if any, has been displayed on the screen
    the cursor has been set to the end of the text
    text_font_style = current text font style
    controlling and keyboard callbacks have been set up for win
    mousing callbacks have been set up for win
    
  test:
    frame maximized on previous run
    frame midimized on previous run
      persisting.get_current_file_name(my_persistents) == null
      persisting.get_current_file_name(my_persistents) != null
  test:
    file_name = None
    file_name = valid file
  """
  global curpos
  global file_name
  global is_loaded
  global text_font_style
  global my_page
  global my_persistents
  wb = window_bounding.get_bounds(win)
  pwb = persisting.get_window_bounds(my_persistents)
  window_bounding.set_bounds(win,persisting.get_window_bounds(my_persistents))
  wb2 = window_bounding.get_bounds(win)
  my_page = rendering.new_page(win,persisting.get_paper_width(my_persistents),persisting.get_paper_height(my_persistents),persisting.get_horizontal_indent(my_persistents),persisting.get_vertical_indent(my_persistents),persisting.get_text_font_name(my_persistents),text_font_style,persisting.get_text_font_size(my_persistents),coloring.BLACK)
  if file_name != None:
    load_file(file_name)
  else:
    persisting.set_text_font_name(my_persistents,DEFAULT_FONT_NAME)
    persisting.set_text_font_size(my_persistents,DEFAULT_FONT_SIZE)
    is_loaded = False
    # set cursor start and end to start of text
    rendering.set_cursor_position_to_start(curpos)
    # draw cursor at origin
    cursoring.draw_cursor(win,persisting.get_text_font_size(my_persistents),persisting.get_horizontal_indent(my_persistents),persisting.get_vertical_indent(my_persistents),coloring.BLACK)
  controlling.attach(win,action_key_hit)
  keyboarding.attach(win,character_key_hit)
  ml = MouseListener()
  mousing.attach(ml,win)
  show_splash_screen()


# Callback classes
# ----------------

class MouseListener(mousing.MouseListener):
  
  def double_clicked(self,x,y):
    """
    pre:
      the primary mouse button has been clicked (pressed and released) twice in quick succession (< 0.5 sec) by the user
      (x,y) = position in points at which the mouse pointer was double-clicked, relative to the top left-hand corner of the window's pane, as floats 

    post:
      the action required by the user has been carried out

    test:
    """
    print("Start of double_clicked")


  def mouse_clicked(self,x,y):
    """
    pre:
      the primary mouse button has been clicked (pressed and released) once by the user
      (x,y) = position in points at which the mouse pointer was clicked, 
                relative to the top left-hand corner of the window's pane, as floats
      my_page = rendering.Page on which mouse-hit occurred
      my_text = current texting.Text
      curpos = current position of cursor
    
    post:
      the cursor has been replaced by a thin cursor at (x,y) on the Page
        and the equivalent (lo,cpo) in the Text
      
    test:
      empty text
      text of two lines or more
    """
    global my_text
    global my_page
    global curpos
    # relativize (x,y) to tlh corner of text
    x -= my_page._horizontal_indent
    y -= my_page._vertical_indent
    
    # find nearest glyph interstice as (page_line_offset, page_glyph_offset, xoff, yoff)
    (plo,pgo,xoff,yoff) = rendering.nearest_glyph_interstice_of(x,y,my_page)
    
    # convert (plo,glo) to (line_offset, code_point_offset) in Text
    (lo,cpo) = rendering.text_position_of(plo,pgo,my_page)
    
    # set up curpos for thin cursor at point of mouse hit in Text and Page
    rendering.set_thin_cursor_text_and_page(curpos,lo,cpo,xoff,yoff)
    # allow lateral movement
    curpos.update_x_offset = True
    # render the thin cursor
    with my_text._my_lock:
      texting.set_cursor(my_text,lo,cpo)
    rendering.render_lines(my_text,curpos,my_page,0,rendering.TO_END)
    rendering.render_thin_cursor(curpos,my_page)
    
    
  def mouse_dragged(self,x,y):
    """
    pre:
      the mouse has been dragged (moved with the primary mouse-button pressed)
        at least 5 points by the user from its previous position
      (x,y) = position in points to which the mouse pointer was dragged, 
        relative to the top left-hand corner of the window's pane, as floats 
      pivot = start position in Text and Page of this mouse-drag operation
      my_text = current Text being edited
      my_page = current Page being rendered
      my_window = current main window
      curpos = current cursor position
    
    post:
      curpos has been set to the section of text between the pivot and 
        the current mouse position
      curpos._start_text_position <= curpos._end_text_position
    
    test:
    """
    global curpos
    global my_page
    global my_persistents
    global my_text
    global my_window
    global pivot
    x_in_text = x - persisting.get_horizontal_indent(my_persistents)
    y_in_text = y - persisting.get_vertical_indent(my_persistents)
    (lo,go,xoff,yoff) = rendering.nearest_glyph_interstice_of(x_in_text,y_in_text,my_page)
    (tlo,cpo) = rendering.text_position_of(lo,go,my_page)
    mouse_position = rendering.new_cursor_position()
    rendering.set_thin_cursor_text_and_page(mouse_position,tlo,cpo,xoff,yoff)
    if rendering._less_than_text_position(pivot._start_text_position,mouse_position._start_text_position):
      curpos._start_text_position.line_offset = \
        pivot._start_text_position.line_offset
      curpos._start_text_position.code_point_offset = \
        pivot._start_text_position.code_point_offset
      curpos._start_page_position.x_offset = pivot._start_page_position.x_offset
      curpos._start_page_position.y_offset = pivot._start_page_position.y_offset
      curpos._end_text_position.line_offset = mouse_position._start_text_position.line_offset
      curpos._end_text_position.code_point_offset = mouse_position._start_text_position.code_point_offset
      curpos._end_page_position.x_offset = mouse_position._start_page_position.x_offset
      curpos._end_page_position.y_offset = mouse_position._start_page_position.y_offset
    else:
      curpos._start_text_position.line_offset = mouse_position._start_text_position.line_offset
      curpos._start_text_position.code_point_offset = mouse_position._start_text_position.code_point_offset
      curpos._start_page_position.x_offset = mouse_position._start_page_position.x_offset
      curpos._start_page_position.y_offset = mouse_position._start_page_position.y_offset
      curpos._end_text_position.line_offset = pivot._start_text_position.line_offset
      curpos._end_text_position.code_point_offset = pivot._start_text_position.code_point_offset
      curpos._end_page_position.x_offset = pivot._start_page_position.x_offset
      curpos._end_page_position.y_offset = pivot._start_page_position.y_offset
    painting.clear_rectangles(my_window)  # get rid of old fat cursor
    rendering.render_lines(my_text,curpos,my_page,0,rendering.TO_END)

    
  def mouse_popup(self,x,y,win,window_x,window_y):
    """
    pre:
      the user has gestured that a popup menu should be displayed. On some platforms, this is done by clicking the right mouse button.
      (x,y) = position in points at which the mouse pointer was clicked, relative to the top left-hand corner of the window's pane, as floats
      win = windowing.Window in which the gesture was made
      (window_x,window_y) = position in points required for popup menu, relative to the top left-hand corner of the window's contents, as floats 
    
    post:
      the action (if any) required by the user has been carried out
    
    test:
    """
    print("Start of mouse_popup")


  def mouse_pressed(self,x,y):
    """
    pre:
      the user has pressed the primary mouse-button, signifying the start of a mouse-drag operation
      (x,y) = position in points of the mouse pointer when the mouse button was pressed, relative to the top left-hand corner of the window's pane, as floats 
      my_persistents = persistent variables for the Emily app
      my_page = current page
  
    post: 
      the start and end points of the cursor have been set to the position
        in the Text and Page equivalent to the mouse_pressed position
      the pivot for the mouse-drag has been set to the position
        in the Text and Page equivalent to the mouse_pressed position
      
    test:
      once thru (check pivot)
    """
    global curpos
    global my_page
    global my_persistents
    global pivot
    x_in_text = x - persisting.get_horizontal_indent(my_persistents)
    y_in_text = y - persisting.get_vertical_indent(my_persistents)
    (lo,go,xoff,yoff) = rendering.nearest_glyph_interstice_of(x_in_text,y_in_text,my_page)
    (tlo,cpo) = rendering.text_position_of(lo,go,my_page)
    # set the cursor start and end
    rendering.set_thin_cursor_text_and_page(curpos,tlo,cpo,xoff,yoff)
    # set the pivot start and end
    rendering.set_thin_cursor_text_and_page(pivot,tlo,cpo,xoff,yoff)
    rendering.render_thin_cursor(curpos,my_page)

  def mouse_released(self,x,y):
    """
    pre:
      the user has released the primary mouse-button, signifying the end of a mouse-drag operation
      (x,y) = position in points of the mouse pointer when the mouse button was released, relative to the top left-hand corner of the window's pane, as floats 
    
    post:
      the action (if any) required by the user has been carried out
    
    test:
    """
    print("Start of mouse_released")

  
# sub-procedures
# --------------

def add_font_menu_item(fs,s,m):
  """ 
  pre:
    fs = font size for this menu item
    s = label of this font menu item
    m = menu to which this menu item is to be added
    my_persistents = persistent variables for the Emily app
  
  post:
    specified font menu item has been added to m

  test:
    s == persisting.get_text_font_name(my_persistents)
    s != persisting.get_text_font_name(my_persistents)
  """
  global my_persistents
  label = build_tagged_label(s, s == persisting.get_text_font_name(my_persistents))
  menuing.add_menu_item(m,fs,label,font_name_item_hit)
  

def build_menu_bar(win):
  """ 
  pre:
    my_persistents = persistent variables for the Emily app
    win = window on which this menu bar is to be built
  
  post:
    menu bar has been set up on win

  test:
    once thru
  """
  global my_persistents
  menuing.add_menu_bar_item(win,persisting.get_menu_font_size(my_persistents),"File",file_bar_item_hit)
  menuing.add_menu_bar_item(win,persisting.get_menu_font_size(my_persistents),"Font",font_bar_item_hit)
  menuing.add_menu_bar_item(win,persisting.get_menu_font_size(my_persistents),"Alignment",alignment_bar_item_hit)
  menuing.add_menu_bar_item(win,persisting.get_menu_font_size(my_persistents),"Setup",setup_bar_item_hit)
  

def build_menu_font_size_menu(fs):
  """ 
  pre:
    fs = font size with which menu is to be built, as int
    my_window = current window for this app
  
  post:
    MenuFontSizeMenu has been built and returned

  test:
    once thru with valid fs
  """
  MINIMUM_FONT_SIZE = 9
  MAXIMUM_FONT_SIZE = 24
  global my_window
  mfsm = menuing.new_menu(my_window)
  pfs = MINIMUM_FONT_SIZE
  while pfs <= MAXIMUM_FONT_SIZE:
    label = build_tagged_font_size_label(pfs,fs)
    menuing.add_menu_item(mfsm,fs,label,menu_font_size_item_hit)
    pfs += 1
  return mfsm
  

def build_tagged_font_size_label(pfs,tfs):
  """ 
  pre:
    pfs = point font size to be used in label, as int
    tfs = point font size to be tagged, as int

  post:
    appropriate label has been returned, as str

  test:
    pfs == tfs
    pfs != tfs
  """
  label = "" + str(pfs) + " point"
  return build_tagged_label(label,pfs == tfs)
  

def build_tagged_label(l,b):
  """ 
  pre:
    l = raw label to be possibly tagged, as str
    b = True, iff l is to be tagged
    
  post:
    appropriate label has been returned, as str

  test:
    b = true
    b = false
  """
  label = ""
  if b:
    label +="• "
    # unicode Bullet Point
  else:
    label += "   "
  label += l
  return label
  

def build_u_i():
  """ 
  pre:  
    my_persistents = persistent variables for the Emily app
    my_window = current window for this app

  post:
    my_window has been set up and displayed
    the zoom factor of the window has been set
    is_loaded = false

  test:
    once thru
  """
  global is_loaded
  global my_persistents
  global my_window
  zoom_factor = persisting.get_zoom_factor(my_persistents)
  my_window = windowing.new_window(persisting.get_menu_font_size(my_persistents),title_of(None),persisting.get_paper_width(my_persistents),persisting.get_paper_height(my_persistents),1.0)
  windowing.set_zoom_factor(my_window,zoom_factor)
  build_menu_bar(my_window)
  is_loaded = False
  windowing.show(my_window,window_opening,window_closing)
  
  
def check_file_type(f):
  """
  pre:
    f = filename to be checked, as str
    my_persistents = persistent variables for the Emily app

  post:
    if f is not a file of type .mle,
      returns None
    else
    returns the filename

  test:
     fred.mlf
     fred.mle
  """
  global my_persistents
  if f[-4:] != ".mle":
    dialoging.show_message_dialog(persisting.get_menu_font_size(my_persistents),"Emily: error in file name", "File name " + str(f) + " is not of type .mle")
    return None
  else:
    return f


def check_save(t):
  """ 
  pre:
    t = text which may need saving
    is_loaded = true, iff current text has been loaded from a file
    my_persistents = persistent variables for the Emily app
  
  post:
    if t has been modified, user has been asked whether they want to save it
      and under what name, if necessary

  test:
    text has not been modified
    text has been modified
      user doesn't want to save
      user wants to save
        text is loaded from file
        text is not loaded from file
  """
  global is_loaded
  global my_persistents
  with t._my_lock:
    text_has_been_modified = texting.has_been_modified(t) 
  if text_has_been_modified:
    user_wants_to_save = dialoging.show_confirm_dialog(persisting.get_menu_font_size(my_persistents),"Text has been modified","The text has been modified.  Do you wish to save it?")
    if user_wants_to_save:
      if is_loaded:
        fn = persisting.get_current_file_name(my_persistents)
        saved_o_k = save_file(fn)
        # if successful, inform user
        if saved_o_k:
          show_successful_save_message(fn)       
      # file is not loaded
      else:
        persisting.set_current_file_name(my_persistents,None)
        save_file_as()
        

def int_of(s):
  """
  pre:
    s = string containing an integer
    
  post:
    the first integer encountered in s has been returned as an int
    
  test:
    s = ""
    s = "fred"
    s = "  90"
    s = "  909 points"
  """
  i = 0
  # find start of integer
  while i < len(s) and (s[i] < '0' or s[i] > '9'):
    i += 1
  if i == len(s):
    raise Exception("int_of cannot find an integer in string \""+ s + "\"")
  #convert integer
  res = 0
  while i < len(s) and s[i] >= '0' and s[i] <= '9':
    res = res*10 + ord(s[i])- ord('0')
    i += 1
  return res


def load_file(s):
  """ 
  pre:
    s = name of .mle file to be loaded, as str
    my_persistents = persistent variables for the Emily app
    my_window = Emily's main window
    my_text = text for this instantiation of Emily
    text_font_style = FontStyles set for this text
    my_page = page to which this file is to be rendered
    
  post:
    Either:
      file s has been loaded as a text in Emily
        and displayed on the screen via my_page
      current_alignment has been set to the alignment of the paragraph containing the cursor
      Emily's title has been updated
      text_font_name and text_font_size have been set up with values from the file s
      is_loaded = true
      File menu has been updated to correspond to is_loaded
      the cursor has been rendered at the end of the text
      curpos = current position of cursor
    or 
      an exception has been thrown and displayed to the user

  test:
    file loads successfully
    file fails to load (e.g. from removed memory stick)
  """
  global curpos
  global current_alignment
  global is_loaded
  global my_persistents
  global my_page
  global my_text
  global my_window
  try:
    empty = ""    # blank current file name
    persisting.set_current_file_name(my_persistents,empty)
    # blank window's title
    windowing.set_title(my_window,title_of(empty))
    la = looking_ahead.lookahead_of(s)
    with my_text._my_lock:
      # parse Emily document to my_text
      r = mle_parsing.parse_mle_document(la)
      # set up new text
      my_text = mle_parsing.text_of(r)
      texting.set_unmodified(my_text)
      # set on-screen cursor's line offset and glyph offset (end of text)
      rendering.set_on_screen_to_text(my_text,curpos)
      # set the current alignment to alignment at cursor position (end of text)
      current_alignment = texting.get_alignment(my_text)
      # set up document's font name and size as persistent variables
      #  and in my_page
      font_name = unicoding3_0.python_string_of(mle_parsing.font_name_of(r))
      persisting.set_text_font_name(my_persistents,font_name)
      rendering.set_text_font_name(my_page,persisting.get_text_font_name(my_persistents))
      persisting.set_text_font_size(my_persistents,mle_parsing.font_size_of(r))
      rendering.set_text_font_size(my_page,persisting.get_text_font_size(my_persistents))
      # clear the window
      windowing.clear(my_window)
      # display text
      texting.set_cursor_start(my_text)
      rendering.render_lines(my_text,curpos,my_page,0,rendering.TO_END)
      # this has set curpos to the end of the text
      #   in the text and on the screen
      rendering.render_thin_cursor(curpos,my_page)
      
    # update current file name
    persisting.set_current_file_name(my_persistents,s)
    # update window's title
    windowing.set_title(my_window,title_of(s))
    # update is_loaded and File menu
    is_loaded = True
    #update_file_menu(persisting.get_menu_font_size(my_persistents),my_file_menu);
    
  except Exception as ex:
    dialoging.show_message_dialog(persisting.get_menu_font_size(my_persistents),"Emily: error while loading input file", str(ex))
    
    
def name_of(s):
  """
  pre:
    s = font name, with possible preceding dots and spaces, as str
    the name in s must start with an alphabetic character,
      but may contain any character thereafter
    
  post:
    font name has been returned
    
  test:
    s = ""
    s = "£££££"
    s = "fredfont"
    s = "& Times New Roman"
  """
  # find name
  i = 0
  while i < len(s) and not s[i].isalpha():
    i += 1
  if i == len(s):
    raise Exception("name_of cannot find a name in string \"" + s + "\"")
  # make result equal to name
  res = ""
  while i < len(s):
    res += s[i]
    i += 1
  return res


def print_cursor(oscp):
  print("OnScreenCursorPosition:")
  print("  oscp._text_position.line_offset="+str(oscp._text_position.line_offset))
  print("  oscp._text_position.code_point_offset="+str(oscp._text_position.code_point_offset))
  print("  oscp._page_position.x_offset="+str(oscp._page_position.x_offset))
  print("  oscp._page_position.y_offset="+str(oscp._page_position.y_offset))
  print("  oscp._desired_x_offset="+str(oscp._desired_x_offset))
  print("  oscp._update_x_offset="+str(oscp._update_x_offset))
  print("  oscp._current_text_position.line_offset="+str(oscp._current_text_position.line_offset))
  print("  oscp._current_text_position.code_point_offset="+str(oscp._current_text_position.code_point_offset))
  print("  oscp._current_page_position.x_offset="+str(oscp._current_page_position.x_offset))
  print("  oscp._current_page_position.y_offset="+str(oscp._current_page_position.y_offset))
  print("  oscp._minimum_page_distance="+str(oscp._minimum_page_distance))    


def print_window_bounds(wb):
  print("window bounds:")
  print("  x="+str(window_bounding.get_x(wb)))
  print("  y="+str(window_bounding.get_y(wb)))
  print("  width="+str(window_bounding.get_width(wb)))
  print("  height="+str(window_bounding.get_height(wb)))
  
  
def render_cursor_left_right():
  """ 
  pre:
    my_text = text to be rendered, with cursor at required position in text
    my_page = page on which text is to be rendered
    curpos = current position of cursor
  
  post:
    text and cursor have been rendered on my_page
    current_alignment = alignment of paragraph containing cursor
  
  note:
    this procedure must be called from inside a block synchronized on my_text

  test:
    once thru
  """
  global curpos
  global current_alignment
  global my_page
  global my_text    
  rendering.render_lines(my_text,curpos,my_page,0,rendering.TO_END)
  # curpos has text and page position set
  rendering.set_text_to_on_screen_start(curpos,my_text)
  current_alignment = texting.get_alignment(my_text)
  rendering.render_thin_cursor(curpos,my_page)
  

def render_modified_text():
  """ 
  pre:
    my_text = text which has been modified
    my_text's cursor is at required position of cursor on screen
    my_page = page to which my_text is to be rendered
    my_page's back map has been set up
    curpos's start page position <= 
      y-offset of start of modified portion of text on page

  post:
    my_text has been rendered to my_page with cursor at required position
    my_text has been marked as modified
    curpos = current cursor position in text and screen

  note:
    this procedure must be called from inside a block synchronized on my_text

  test:
    once thru
  """
  global my_page
  global my_text
  global curpos
  rendering.set_on_screen_to_text(my_text,curpos)
  # curpos has text position set
  start_line = rendering.start_page_line_of(curpos,my_page)
  rendering.render_lines(my_text,curpos,my_page,start_line,rendering.TO_END)
  # curpos has text and page positions set
  rendering.render_thin_cursor(curpos,my_page)
  texting.set_modified(my_text)
  

def save_file(s):
  """ 
  pre:
    s = name of file to be saved
    is_loaded = True, iff loaded file is s
    VERSION = current version of Emily
    my_text = text to be output as contents of file
    my_persistents = persistent variables for the Emily app
  
  post:
    Either:
      my_text has not been modified, and no action has been taken
      True has been returned
    Or:
      specified file has been saved from Emily's text
      persisting.get_current_file_name(my_persistents) = s
      my_text state has been set to unmodified
      is_loaded = True
      True has been returned
    Or: 
      an exception has been thrown and displayed to the user
      False has been returned

  test:
    file is not loaded
    file is loaded
      file saves successfully
      file fails to save (e.g. to removed memory stick)
  """
  global is_loaded
  global my_persistents
  global my_text
  if texting.has_been_modified(my_text):
    if not is_loaded:
      save_file_as()
      return True
    else:
      try:
        w = unicode_io.new_output_writer(s)
        font_name = unicoding3_0.string_of(persisting.get_text_font_name(my_persistents))
        mle_emitting.emit_emily_document(font_name,persisting.get_text_font_size(my_persistents),unicoding3_0.string_of(VERSION),my_text,w)
        # if we make it to here, save was successful
        persisting.set_current_file_name(my_persistents,s)
        # text and file are the same
        with my_text._my_lock:
          texting.set_unmodified(my_text)
        return True
          
      except Exception as ex:
        dialoging.show_message_dialog(persisting.get_menu_font_size(my_persistents),"Exception writing output file",str(ex))
        raise ex
        return False
    
  
def save_file_as():
  """
  pre:
    VERSION = current version of Emily
    my_text = text to be output as contents of file
    my_persistents = persistent variables for the Emily app
    my_window = main window of Emily
    persisting.get_current_file_name(my_persistents) = name to save file as, 
      or null if default directory
  
  post:
    either:
      user has declined to save
    or
      either:
        specified file has been saved from Emily's text
        persisting.get_current_file_name(my_persistents) = name of saved file
        my_text state has been set to unmodified
        is_loaded = true
        File menu has been updated to correspond to is_loaded
        the user has been informed that the file has been successfuly saved
      or 
        an exception has been thrown and displayed to the user
    main window title has been modified to correspond to the current file name

  test:
    user disagrees
    user agrees
      file saved OK
      file saving fails
  """
  global is_loaded
  global my_persistents
  global my_text
  global my_window
  # display file saving dialog
  s = file_dialoging.show_save_file_dialog(persisting.get_menu_font_size(my_persistents),"Save As",persisting.get_current_file_name(my_persistents),["mle"],file_dialoging.SortMode.LATEST_FIRST)
  if s != None:
    is_loaded = True  # to force the save
    with my_text._my_lock:
      texting.set_modified(my_text)  # to force the save
    saved_o_k = save_file(s)
    # if successful, inform user
    if saved_o_k:
      show_successful_save_message(s)
    windowing.set_title(my_window,title_of(persisting.get_current_file_name(my_persistents)))
  else:
    dialoging.show_message_dialog(persisting.get_menu_font_size(my_persistents),"File save aborted","Problem saving file.  File has not been saved.")
    
  
def set_current_alignment(a):
  """ 
  pre:
    a = required current alignment
    curpos = position of cursor on the page
    my_text = text being edited
    my_page = page to which my_text is rendered
  
  post: 
    current_alignment = a
    the alignment of the paragraph containing the cursor in my_text has been set to a
    the text and cursor have been re-rendered on my_page

  test:
    once thru
  """
  global current_alignment
  global my_text
  current_alignment = a
  with my_text._my_lock:
    rendering.set_text_to_on_screen_end(curpos,my_text)
    texting.set_alignment(my_text,a)
    rendering.render_lines(my_text,curpos,my_page,0,rendering.TO_END)
  

""" 
  pre:  my_persistents.get_splash_screen_on() = state of splash_screen_on
  post: if splash_screen_on,
          starts a thread which shows the splash screen for 1 sec.
  
  """
def show_splash_screen():
  global my_persistents
  #if (persisting.get_splash_screen_on(my_persistents))
  #SplashScreenThread sst = new SplashScreenThread();
  #sst.start();
  pass
  
  pass
  

""" 
  pre:
    fn = name of file which has been successfully saved
    my_persistents has the required menu font size
  post:
    "successful save" message has been shown and accepted by the user.
  """
""" 
  test:
    once thru
  """
def show_successful_save_message(fn):
  global my_persistents
  mess = "The file \"" + fn + "\" has been successfully saved"
  dialoging.show_message_dialog(persisting.get_menu_font_size(my_persistents),"File successfully saved",mess)
  

""" 
  pre:
    s = name of currently loaded file, or null if none
    VERSION = current version of Emily
  post:
    returns title to be displayed in Emily's title bar
  """
""" 
  test:
    s is null
    s is non-null
  """
def title_of(s):
  title = "Emily " + VERSION
  if s != None:
    title += " - " + s    
  return title
  

# --------------------------------------------------------------------------
# main program
# --------------------------------------------------------------------------

def _print_pos(curpos):
  print("    curpos._start_text_position.line_offset=" + str(curpos._start_text_position.line_offset))
  print("    curpos._start_text_position.code_point_offset=" + str(curpos._start_text_position.code_point_offset))
  print("    curpos._end_text_position.line_offset=" + str(curpos._end_text_position.line_offset))
  print("    curpos._end_text_position.code_point_offset=" + str(curpos._end_text_position.code_point_offset))
  print("    curpos._start_page_position.x_offset=" + str(curpos._start_page_position.x_offset))
  print("    curpos._start_page_position.y_offset=" + str(curpos._start_page_position.y_offset))
  print("    curpos._end_page_position.x_offset=" + str(curpos._end_page_position.x_offset))
  print("    curpos._end_page_position.y_offset=" + str(curpos._end_page_position.y_offset))
  print("    curpos._desired_x_offset=" + str(curpos._desired_x_offset))
  print("    curpos._update_x_offset=" + str(curpos._update_x_offset))
  print("    curpos._start_x_offset=" + str(curpos._start_x_offset))
  print("    curpos._end_x_offset=" + str(curpos._end_x_offset))
  print("    curpos._in_fat_cursor=" + str(curpos._in_fat_cursor))
  print("    curpos._cursor_rectangle_pending=" + str(curpos._cursor_rectangle_pending))



"""
Start-up program for emily version 0.9
This program is required to ensure emily.py is started as "emily0_9.emily" rather than "__main__", and therefore that relative module addressing works.
"""
# author R.N.Bosworth

# version 17 Mar 23  10:40
"""
Copyright (C) 2023  R.N.Bosworth

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License (gpl.txt) for more details.
"""

import sys
from emily0_9 import emily

emily.start(sys.argv)



# Contractor to parse an HTML string.

# author R.N.Bosworth

# version 4 Mar 23  16:31

from . import looking_ahead

from guibits1_0 import type_checking2_0, unicoding3_0

""" 
Copyright (C) 2019,2021,2022,2023  R.N.Bosworth

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License (gpl.txt) for more details.
"""
""" 
  HTML Grammar version 1 Dec 2018   15:05

  This grammar defines the basic HTML constructs common to any HTML file.

  This grammar is in EBNF (Extended Backus-Naur Form) using the following meta-symbols:  ::==  is  (  )  |  *.  The equivalent terminal characters are called LPARENTHESIS, RPARENTHESIS, BAR and STAR.

  Non-terminals are in CamelCase.  Terminal names are in UPPER_CASE.  Other symbols represent themselves.
  
  The rules are case-sensitive, unlike HTML5.
  
  
  StartTag(Name)  ::==  < Name OptionalSpaces >

  EndTag(Name)  ::==  < / Name OptionalSpaces >

  HtmlCharacter  ::==  ( & a m p ;
                       | & g t ;
                       | & l t ;
                       | any Unicode code point except '&' (U+26), '>' (U+3E) and '<' (U+3C)

Note: the rules for OptionalSpaces, SpaceChar, Separator, Name and NameChar are in the HTML CSS Grammar, as they are common to both HTML and CSS
"""


# exposed procedures
# ------------------

def parse_html_character(s,o):
  """ 
  pre:
    s = unicoding3_0.String to be parsed
    o = int offset from which the translation is to start
    
  post:
    either:
      an (escaped) unicode code point has been consumed from s
      a pair of ints has been returned, consisting of the equivalent non-escaped unicode code point and the offset past the consumed (escaped) unicode code point
    or:
      an exception has been raised
      
  syntax:
    HtmlCharacter  ::==  ( & a m p ;
                         | & g t ;
                         | & l t ;
                         | any Unicode code point except '&' (U+26), '>' (U+3E) and '<' (U+3C)

  test:
    l = None
        "&bmp;"
        "&anp;"
        "&amq;"
        "&amp;"
        "&ft;"
        "&gs;"
        "&gt;"
        "&mt;"
        "&lu;"
        "&lt;"
        ">"
        "<"
        "a"
  """
  type_checking2_0.check_derivative(s,unicoding3_0.String)
  type_checking2_0.check_identical(o,int)
  cp = unicoding3_0.code_point_at(s,o)
  if cp == ord('&'):
    o += 1
    cp = unicoding3_0.code_point_at(s,o)
    if cp == ord('a'):
      o = _accept(ord('a'),s,o)
      o = _accept(ord('m'),s,o)
      o = _accept(ord('p'),s,o)
      o = _accept(ord(';'),s,o)
      return (ord('&'),o)
      
    elif cp == ord('g'):
      o = _accept(ord('g'),s,o)
      o = _accept(ord('t'),s,o)
      o = _accept(ord(';'),s,o)
      return (ord('>'),o)
      
    elif cp == ord('l'):
      o = _accept(ord('l'),s,o)
      o = _accept(ord('t'),s,o)
      o = _accept(ord(';'),s,o)
      return (ord('<'),o)
      
    else:
      raise Exception("unexpected code point in HTML escape: U+" + format(cp,'x'))
      
  elif cp == ord('>'):
    raise Exception("illegal code point in text: U+" + format(ord('>'),'x'))
    
  elif cp == ord('<'):
    raise Exception("illegal code point in text: U+" + format(ord('<'),'x'))
    
  else:
    o += 1
    return (cp,o)
    

# private procedures
# ------------------

def _accept(cp,s,o):
  """ 
  pre:
    cp = unicode code point which is to be accepted, as an int
    s = unicoding3_0.String which is being parsed
    o = offset in s of code point to be tested
    
  post:
    either:
      cp has been accepted and consumed from s
      o has been updated past the accepted code point
    or:
      an exception has been raised

  test:
    cp = 'A'
      l = "B"
      l = "A"
  """
  if unicoding3_0.code_point_at(s,o) != cp:
    raise Exception("unexpected code point, expected: U+" + format(cp,'x') + ", found: U+" + format(unicoding3_0.code_point_at(s,o),'x'))
  else:
    o += 1
    return o



# Test program for Contractor which has procedures for parsing basic html entities.

# author R.N.Bosworth

# version 9 Mar 23  14:26

from emily0_9 import html_parsing, looking_ahead
from guibits1_0 import unicoding3_0

""" 
Copyright (C) 2017,2018,2019,2021,2022,2023  R.N.Bosworth

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License (gpl.txt) for more details.
"""

# test program
# ------------

print("Tests of html_parsing.accept_end_tag", end=' ')
#html_parsing.accept_end_tag(unicoding3_0.string_of("html"),None)
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of(">/html>"))
#html_parsing.accept_end_tag(None,la)
#html_parsing.accept_end_tag(unicoding3_0.string_of("html"),la)
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("<\\html>"))
#html_parsing.accept_end_tag(unicoding3_0.string_of("html"),la)
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("</html<"))
#html_parsing.accept_end_tag(unicoding3_0.string_of("html"),la)
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("</htmm>"))
#html_parsing.accept_end_tag(unicoding3_0.string_of("html"),la)
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("</html>"))
html_parsing.accept_end_tag(unicoding3_0.string_of("html"),la)
assert looking_ahead.current_symbol_of(la) == looking_ahead.END_OF_STREAM
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("</HTML  >"))
#html_parsing.accept_end_tag(unicoding3_0.string_of("html"),la)
print("OK")

print("Tests of accept_start_tag", end=' ')
#html_parsing.accept_start_tag(unicoding3_0.string_of("html"),None)
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of(">html>"))
#html_parsing.accept_start_tag(None,la)
#html_parsing.accept_start_tag(unicoding3_0.string_of("html"),la)
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("<html<"))
#html_parsing.accept_start_tag(unicoding3_0.string_of("html"),la)
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("<htmm>"))
#html_parsing.accept_start_tag(unicoding3_0.string_of("html"),la)
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("<html>"))
html_parsing.accept_start_tag(unicoding3_0.string_of("html"),la)
assert looking_ahead.current_symbol_of(la) == looking_ahead.END_OF_STREAM
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("<HTML  >"))
#html_parsing.accept_start_tag(unicoding3_0.string_of("html"),la)
print("OK")

print("Tests of parse_start_tag", end=' ')
#s = html_parsing.parse_start_tag(None)
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("{fred  }"))
#s = html_parsing.parse_start_tag(la)
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("<  >"))
#s = html_parsing.parse_start_tag(la)
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("<fred  >"))
s = html_parsing.parse_start_tag(la)
assert unicoding3_0.equals(s,unicoding3_0.string_of("fred"))
assert looking_ahead.current_symbol_of(la) == looking_ahead.END_OF_STREAM
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("<jim]"))
#s = html_parsing.parse_start_tag(la)
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("<jim>"))
s = html_parsing.parse_start_tag(la)
assert unicoding3_0.equals(s,unicoding3_0.string_of("jim"))
assert looking_ahead.current_symbol_of(la) == looking_ahead.END_OF_STREAM
print("OK")

print("Tests of parse_tag", end=' ')
#s = html_parsing.parse_tag(None)
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("{/fred  }"))
#s = html_parsing.parse_tag(la)
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("</  >"))
#s = html_parsing.parse_tag(la)
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("</fred  >"))
s = html_parsing.parse_tag(la)
assert unicoding3_0.equals(s,unicoding3_0.string_of("/fred"))
assert looking_ahead.current_symbol_of(la) == looking_ahead.END_OF_STREAM
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("<jim]"))
#s = html_parsing.parse_tag(la)
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("<jim>"))
s = html_parsing.parse_tag(la)
assert unicoding3_0.equals(s,unicoding3_0.string_of("jim"))
assert looking_ahead.current_symbol_of(la) == looking_ahead.END_OF_STREAM
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("</bert>"))
s = html_parsing.parse_tag(la)
assert unicoding3_0.equals(s,unicoding3_0.string_of("/bert"))
assert looking_ahead.current_symbol_of(la) == looking_ahead.END_OF_STREAM
print("OK")

print("")
print("All tests OK")



import io
import json
import os
from guibits1_0 import type_checking2_0, window_bounding

# author R.N.Bosworth

# version 25 Feb 2023  19:09
""" 
Contractor which deals with persistent data for Emily.

Note: test version with tiny page!

Copyright (C) 2016,2017,2018,2021,2022,2023  R.N.Bosworth

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License (gpl.txt) for more details.
"""

# exposed types
# -------------

class Persistents:  # container object for persistent data
   _data = {"current_file_name":None,"menu_font_size":16.0, \
            "splash_screen_on":False,"text_font_name":"Times New Roman", \
            "text_font_size":12.0, \
            "window_bounds":{"x":72.0,"y":72.0,"width":576.0,"height":288.0}, \
            "zoom_factor":1.0}
  
  
# exposed procedures
# ------------------

def get_current_file_name(p):
  """ 
  pre:
    p = Persistents container for required persistent value
    
  post:
    current file name has been returned as str, or None if no current file name

  test:
    file name None
    file name valid
  """
  type_checking2_0.check_derivative(p,Persistents)
  return p._data["current_file_name"]


def get_persistents():
  """ 
  pre:
    true
    
  post:
    a Persistents object containing persistent values
      has been returned
    either:
      the persistents have been read in from the "EmilyPersistents.json" file in the 
        user's default directory
    or:
      it was not possible to read the "EmilyPersistents.json" file
      a default Persistents has been returned
      
  note:
    This object is a singleton.
 
  test:
    _the_persistents  is None
    _the_persistents  is valid
  """
  global _the_persistents
  if _the_persistents == None:
    try:
      # set up defaults
      _the_persistents = Persistents()
      default_directory = os.path.expanduser("~")  # user's default directory
      save_file = open(os.path.join(default_directory,"EmilyPersistents.json"))
      _the_persistents._data = json.load(save_file)
    except Exception:
      pass  # the file was not present
  return _the_persistents
  

def get_horizontal_indent(p):
  """ 
  pre:
    p = Persistents container for required persistent value
    
  post:
    horizontal indent in points has been returned, as float
    
  test:
    once thru
  """
  type_checking2_0.check_derivative(p,Persistents)
  return 72.0
  

def get_paper_height(p):
  """ 
  pre:
    p = Persistents container for required persistent value
  post:
    height of paper in points has been returned, as float
    
  Note: test version with tiny page!

  test:
    once thru
  """
  type_checking2_0.check_derivative(p,Persistents)
  return 841.0
  # slightly smaller to prevent printer probs
  # return 216.0;   // i.e. 1" of print for testing purposes
  

def get_paper_width(p):
  """ 
  pre:
    p = Persistents container for required persistent value
    
  post:
    width of paper in points has been returned, as float
    
  Note: test version with tiny page!

  test:
    once thru
  """
  type_checking2_0.check_derivative(p,Persistents)
  return 592.0
  # slightly smaller to prevent printer probs
  #return 288.0;  // i.e. 2 inches of print for testing purposes
  

def get_menu_font_size(p):
  """ 
  pre:
    p = Persistents container for required persistent value
    
  post:
    value of menu_font_size has been returned, as float

  test:
    once thru
  """
  type_checking2_0.check_derivative(p,Persistents)
  return p._data["menu_font_size"]
  

def get_splash_screen_on(p):
  """ 
  pre:
    p = Persistents container for required persistent value
    
  post:
    value of splash_screen_on has been returned, as bool

  test:
    once thru
  """
  type_checking2_0.check_derivative(p,Persistents)
  return p._data["splash_screen_on"]
  

def get_text_font_name(p):
  """ 
  pre:
    p = Persistents container for required persistent value
    
  post:
    value of text_font_name has been returned, as str

  test:
    once thru
  """
  type_checking2_0.check_derivative(p,Persistents)
  return p._data["text_font_name"]
  

def get_text_font_size(p):
  """ 
  pre:
    p = Persistents container for required persistent value
    
  post:
    value of font_text_size has been returned, as float

  test:
    once thru
  """
  type_checking2_0.check_derivative(p,Persistents)
  return p._data["text_font_size"]
  

def get_vertical_indent(p):
  """ 
  pre:
    p = Persistents container for required persistent value
    
  post:
    vertical indent in points has been returned, as float

  test:
    once thru
  """
  type_checking2_0.check_derivative(p,Persistents)
  return 72.0
  

def get_window_bounds(p):
  """ 
  pre:
    p = Persistents container for required persistent value
    
  post:
    value of window_bounds has been returned, as a window_bounding.WindowBounds

  test:
    once thru with non-zero bounds
  """
  type_checking2_0.check_derivative(p,Persistents) 
  wbd = p._data["window_bounds"]  # window bounds as dict
  wb = window_bounding.new_bounds()
  window_bounding.set_x(wb,wbd["x"])
  window_bounding.set_y(wb,wbd["y"])
  window_bounding.set_width(wb,wbd["width"])
  window_bounding.set_height(wb,wbd["height"])
  return wb
  

def get_zoom_factor(p):
  """ 
  pre:
    p = Persistents container for required persistent value
    
  post:
    value of zoom factor has been returned, as float

  test:
    once thru
  """
  type_checking2_0.check_derivative(p,Persistents)  
  return p._data["zoom_factor"]
  

def set_current_file_name(p,s):
  """ 
  pre:
    p = Persistents container for required persistent value
    s = file name to which current_file_name is to be updated, as str
          or None
          
  post:
    current_file_name has been updated to required value
    Persistents have been saved to disk

  test:
    s  is null
    s  is non-null
  """
  type_checking2_0.check_derivative(p,Persistents)
  if s != None:
    type_checking2_0.check_identical(s,str)
  p._data["current_file_name"] = s
  _save_persistents()
  

def set_menu_font_size(p,d):
  """ 
  pre:
    p = Persistents container for required persistent value
    d = size to which menu_font_size is to be updated
    
  post:
    menu_font_size has been updated to required value
    Persistents have been saved to disk

  test:
    once thru
  """
  type_checking2_0.check_derivative(p,Persistents)
  type_checking2_0.check_identical(d,float)  
  p._data["menu_font_size"] = d
  _save_persistents()
  

def set_splash_screen_on(p,b):
  """ 
  pre:
    p = Persistents container for required persistent value
    b = true iff splash_screen is on
    
  post:
    splash_screen_on has been updated to required value
    Persistents have been saved to disk

  test:
    once thru
  """
  type_checking2_0.check_derivative(p,Persistents)
  type_checking2_0.check_derivative(b,bool)
  p._data["splash_screen_on"] = b
  _save_persistents()
  

def set_text_font_name(p,s):
  """ 
  pre:
    p = Persistents container for required persistent value
    s = name to which text_font_name is to be updated
    
  post:
    text_font_name has been updated to required value
    Persistents have been saved to disk

  test:
    once thru
  """
  type_checking2_0.check_derivative(p,Persistents)
  type_checking2_0.check_identical(s,str)
  p._data["text_font_name"] = s
  _save_persistents()
  

def set_text_font_size(p,d):
  """ 
  pre:
    p = Persistents container for required persistent value
    d = size to which text_font_size is to be updated
    
  post:
    text_font_size has been updated to required value
    Persistents have been saved to disk

  test:
    once thru
  """
  type_checking2_0.check_derivative(p,Persistents)
  type_checking2_0.check_identical(d,float)
  p._data["text_font_size"] = d
  _save_persistents()
  

  """ 
  pre:
    p = Persistents container for required persistent value
    wb = window_bounds which are to be set
  post:
    window_bounds has been updated to required value
    Persistents have been saved to disk
  """
""" 
  pre:
    _the_persistents = p
  """
""" 
  test:
    once thru
  """


def set_window_bounds(p,wb):
  type_checking2_0.check_derivative(p,Persistents)
  type_checking2_0.check_derivative(wb,window_bounding.WindowBounds)
  wb_dict = {"x":window_bounding.get_x(wb),"y":window_bounding.get_y(wb), \
             "width":window_bounding.get_width(wb),"height":window_bounding.get_height(wb)}
  p._data["window_bounds"] = wb_dict
  _save_persistents()
  

  """ 
  pre:
    p = Persistents container for required persistent value
    d = value to which zoom factor is to be updated
    
  post:
    zoom factor has been updated to required value
    Persistents have been saved to disk

  test:
    once thru
  """


def set_zoom_factor(p,d):
  type_checking2_0.check_derivative(p,Persistents)
  type_checking2_0.check_identical(d,float)
  p._data["zoom_factor"] = d
  _save_persistents()
  

# private members
# ---------------

_the_persistents = None
  

def _save_persistents():
  """ 
  pre:
    _the_persistents = persistents object to  be saved
    
  post:
    an attempt has been made to save _the_persistents in user's default directory
      under the name EmilyPersistents.json

  test:
    once thru
  """
  default_directory = os.path.expanduser("~")  # user's default directory
  save_file = open(os.path.join(default_directory,"EmilyPersistents.json"),'w')
  try:
    json.dump(_the_persistents._data,save_file)
  except:
    print("Failed to save persistents")
  


# Test program for Contractor for parsing Cascading Style Sheets (simplified syntax!).

# author R.N.Bosworth

# version 9 Mar 2023  15:43

from emily0_9 import css_parsing, looking_ahead
from guibits1_0 import unicoding3_0

""" 
Contractor for parsing Cascading Style Sheets (simplified syntax!).

Copyright (C) 2016,2017,2018,2021,2022,2023  R.N.Bosworth

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License (gpl.txt) for more details.
"""

# test program
# ------------

print("Tests of css_parsing._new_style_properties", end=' ')
sp = css_parsing._new_style_properties()
assert len(sp._text_alignment_map) == 0
assert sp._font_name == None
assert sp._font_size == 0.0
print("OK")

print("Tests of css_parsing._parse_dot_name", end=' ')
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("%"))
#css_parsing._parse_dot_name(la);
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("f"))
assert unicoding3_0.equals(css_parsing._parse_dot_name(la),unicoding3_0.string_of("f"))
assert looking_ahead.current_symbol_of(la) == looking_ahead.END_OF_STREAM
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("fred%"))
assert unicoding3_0.equals(css_parsing._parse_dot_name(la),unicoding3_0.string_of("fred"))
assert looking_ahead.current_symbol_of(la) == ord('%')
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("fred.jim"))
assert unicoding3_0.equals(css_parsing._parse_dot_name(la),unicoding3_0.string_of("fred.jim"))
assert looking_ahead.current_symbol_of(la) == looking_ahead.END_OF_STREAM
print("OK")

print("Tests of css_parsing._parse_text_align_value", end=' ')
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("bollocks"))
#s = css_parsing._parse_text_align_value(la);
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("left"))
s = css_parsing._parse_text_align_value(la)
assert unicoding3_0.equals(s,unicoding3_0.string_of("left"))
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("LEFT"))
#s = css_parsing._parse_text_align_value(la);
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("RIGHT"))
#s = css_parsing._parse_text_align_value(la);
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("right"))
s = css_parsing._parse_text_align_value(la)
assert unicoding3_0.equals(s,unicoding3_0.string_of("right"))
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("center"))
s = css_parsing._parse_text_align_value(la)
assert unicoding3_0.equals(s,unicoding3_0.string_of("center"))
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("CENTER"))
#s = css_parsing._parse_text_align_value(la);
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("JUSTify"))
#s = css_parsing._parse_text_align_value(la);
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("just_i_f_y"))
#s = css_parsing._parse_text_align_value(la);
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("justify"))
s = css_parsing._parse_text_align_value(la)
assert unicoding3_0.equals(s,unicoding3_0.string_of("justify"))
print("OK")

print("Tests of css_parsing._parse_text_alignment_declaration", end=' ')
m = dict()
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("text-aligm:left"))
#css_parsing._parse_text_alignment_declaration(la,unicoding3_0.string_of("fred.begin"),m);
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("text-align:left"))
css_parsing._parse_text_alignment_declaration(la,unicoding3_0.string_of("fred.begin"),m)
assert m["fred.begin"] == "left"
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("text-align   :   justify"))
css_parsing._parse_text_alignment_declaration(la,unicoding3_0.string_of("jim.just"),m)
assert m["jim.just"] == "justify"
print("OK")

print("Tests of css_parsing._is_font_name_char", end=' ')
assert  not css_parsing._is_font_name_char(ord('@'))
assert css_parsing._is_font_name_char(ord('A'))
assert css_parsing._is_font_name_char(ord('Z'))
assert  not css_parsing._is_font_name_char(ord('['))
assert  not css_parsing._is_font_name_char(ord('`'))
assert css_parsing._is_font_name_char(ord('a'))
assert css_parsing._is_font_name_char(ord('z'))
assert  not css_parsing._is_font_name_char(ord('{'))
assert  not css_parsing._is_font_name_char(ord(''))
assert css_parsing._is_font_name_char(ord(' '))
assert  not css_parsing._is_font_name_char(ord('!'))
print("OK")

print("Tests of css_parsing._parse_font_name", end=' ')
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("\'\'"))
#s = css_parsing._parse_font_name(la);
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("\"\""))
#s = css_parsing._parse_font_name(la);
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("\"a\""))
s = css_parsing._parse_font_name(la)
assert unicoding3_0.equals(s,unicoding3_0.string_of("a"))
assert looking_ahead.current_symbol_of(la) == looking_ahead.END_OF_STREAM
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("\"a%\""))
#s = css_parsing._parse_font_name(la);
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("\"Times Roman\'"))
#s = css_parsing._parse_font_name(la);
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("\"Times Roman\""))
s = css_parsing._parse_font_name(la)
assert unicoding3_0.equals(s,unicoding3_0.string_of("Times Roman"))
assert looking_ahead.current_symbol_of(la) == looking_ahead.END_OF_STREAM
print("OK")

print("Tests of css_parsing._parse_font_name_declaration_rest", end=' ')
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("gamily:\"New Century Schoolbook\""))
#s = css_parsing._parse_font_name_declaration_rest(la);
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("family;\"New Century Schoolbook\""))
#s = css_parsing._parse_font_name_declaration_rest(la);
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("family:\"New Century Schoolbook\""))
s = css_parsing._parse_font_name_declaration_rest(la)
assert unicoding3_0.equals(s,unicoding3_0.string_of("New Century Schoolbook"))
assert looking_ahead.current_symbol_of(la) == looking_ahead.END_OF_STREAM
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("family  :  \"Gill Sans\""))
s = css_parsing._parse_font_name_declaration_rest(la)
assert unicoding3_0.equals(s,unicoding3_0.string_of("Gill Sans"))
assert looking_ahead.current_symbol_of(la) == looking_ahead.END_OF_STREAM
print("OK")

print("Tests of css_parsing._is_digit", end=' ')
assert  not css_parsing._is_digit(ord('/'))
assert css_parsing._is_digit(ord('0'))
assert css_parsing._is_digit(ord('9'))
assert  not css_parsing._is_digit(ord(':'))
print("OK")

print("Tests of css_parsing._parse_font_size_declaration_rest", end=' ')
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("tize:0pt"))
#d = css_parsing._parse_font_size_declaration_rest(la);
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("size;0pt"))
#d = css_parsing._parse_font_size_declaration_rest(la);
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("size:/pt"))
#d = css_parsing._parse_font_size_declaration_rest(la);
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("size:0pt"))
d = css_parsing._parse_font_size_declaration_rest(la)
assert d == 0.0
assert looking_ahead.current_symbol_of(la) == looking_ahead.END_OF_STREAM
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("size:9pt"))
d = css_parsing._parse_font_size_declaration_rest(la)
assert d == 9.0
assert looking_ahead.current_symbol_of(la) == looking_ahead.END_OF_STREAM
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("size::pt"))
#d = css_parsing._parse_font_size_declaration_rest(la);
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("size:1/pt"))
#d = css_parsing._parse_font_size_declaration_rest(la);
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("size:10pt"))
d = css_parsing._parse_font_size_declaration_rest(la)
assert d == 10.0
assert looking_ahead.current_symbol_of(la) == looking_ahead.END_OF_STREAM
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("size:19pt"))
d = css_parsing._parse_font_size_declaration_rest(la)
assert d == 19.0
assert looking_ahead.current_symbol_of(la) == looking_ahead.END_OF_STREAM
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("size:1:pt"))
#d = css_parsing._parse_font_size_declaration_rest(la);
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("size:99qt"))
#d = css_parsing._parse_font_size_declaration_rest(la);
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("size:99pu"))
#d = css_parsing._parse_font_size_declaration_rest(la);
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("size:99pt"))
d = css_parsing._parse_font_size_declaration_rest(la)
assert d == 99.0
assert looking_ahead.current_symbol_of(la) == looking_ahead.END_OF_STREAM
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("size  :  99pt"))
d = css_parsing._parse_font_size_declaration_rest(la)
assert d == 99.0
assert looking_ahead.current_symbol_of(la) == looking_ahead.END_OF_STREAM
print("OK")

print("Tests of css_parsing._parse_font_declaration", end=' ')
sp = css_parsing._new_style_properties()
assert sp._font_name == None
assert sp._font_size == 0.0
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("gont-family : \"Times New Roman\""))
#css_parsing._parse_font_declaration(la,sp);
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("font=family : \"Times New Roman\""))
#css_parsing._parse_font_declaration(la,sp);
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("font-family : \"Times New Roman\""))
css_parsing._parse_font_declaration(la,sp)
assert unicoding3_0.equals(sp._font_name,unicoding3_0.string_of("Times New Roman"))
assert sp._font_size == 0.0
assert looking_ahead.current_symbol_of(la) == looking_ahead.END_OF_STREAM
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("font-size : 42pt"))
css_parsing._parse_font_declaration(la,sp)
assert unicoding3_0.equals(sp._font_name,unicoding3_0.string_of("Times New Roman"))
assert sp._font_size == 42.0
assert looking_ahead.current_symbol_of(la) == looking_ahead.END_OF_STREAM
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("font-blob : \"blobbie\""))
#css_parsing._parse_font_declaration(la,sp);
print("OK")

print("Tests of css_parsing._parse_declaration", end=' ')
sp = css_parsing._new_style_properties()
assert sp._font_name == None
assert sp._font_size == 0.0
assert len(sp._text_alignment_map) == 0
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("font-family : \"Times New Roman\""))
css_parsing._parse_declaration(la,sp,unicoding3_0.string_of("body"))
assert unicoding3_0.equals(sp._font_name,unicoding3_0.string_of("Times New Roman"))
assert sp._font_size == 0.0
assert len(sp._text_alignment_map) == 0
assert looking_ahead.current_symbol_of(la) == looking_ahead.END_OF_STREAM
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("text-align:left"))
css_parsing._parse_declaration(la,sp,unicoding3_0.string_of("p.begin"))
assert unicoding3_0.equals(sp._font_name,unicoding3_0.string_of("Times New Roman"))
assert sp._font_size == 0.0
assert len(sp._text_alignment_map) == 1
assert sp._text_alignment_map["p.begin"] == "left"
assert looking_ahead.current_symbol_of(la) == looking_ahead.END_OF_STREAM
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("complete-load-of-rubbish"))
#css_parsing._parse_declaration(la,sp,unicoding3_0.string_of("p.end"));
print("OK")

print("Tests of css_parsing._parse_declaration_block", end=' ')
sp = css_parsing._new_style_properties()
s = unicoding3_0.string_of("fred")
assert sp._font_name == None
assert sp._font_size == 0.0
assert len(sp._text_alignment_map) == 0
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("[  font-family:\"Times New Roman\"}"))
#css_parsing._parse_declaration_block(la,sp,s);
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("{  zont-family:\"Times New Roman\"}"))
#css_parsing._parse_declaration_block(la,sp,s);
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("{  font-family:\"Times New Roman\"  ]"))
#css_parsing._parse_declaration_block(la,sp,s);
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("{  font-family:\"Times New Roman\"  }"))
#css_parsing._parse_declaration_block(la,sp,s);
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("{font-family:\"Times New Roman\";}"))
css_parsing._parse_declaration_block(la,sp,s)
assert unicoding3_0.equals(sp._font_name,unicoding3_0.string_of("Times New Roman"))
assert sp._font_size == 0.0
assert len(sp._text_alignment_map) == 0
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("{  font-family:\"Times New Roman\"  ;   font-family:\"Palatino\"  }"))
#css_parsing._parse_declaration_block(la,sp,s);
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("{  font-family:\"Times New Roman\"  ;   font-family:\"Palatino\"  ;  }"))
css_parsing._parse_declaration_block(la,sp,s)
assert unicoding3_0.equals(sp._font_name,unicoding3_0.string_of("Palatino"))
assert sp._font_size == 0.0
assert len(sp._text_alignment_map) == 0
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("" + "{\n" + "  font-family:\"Times New Roman\";\n" + "  font-size:10pt;\n" + "  font-family:\"Palatino\";\n" + "  font-size:20pt;\n" + "  text-align:left;\n" + "  text-align:center;\n" + "}"))
css_parsing._parse_declaration_block(la,sp,s)
assert unicoding3_0.equals(sp._font_name,unicoding3_0.string_of("Palatino"))
assert sp._font_size == 20.0
assert len(sp._text_alignment_map) == 1
assert sp._text_alignment_map["fred"] == "center"
print("OK")

print("Tests of css_parsing._parse_rule", end=' ')
sp = css_parsing._new_style_properties()
assert sp._font_name == None
assert sp._font_size == 0.0
assert len(sp._text_alignment_map) == 0
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("p.begin   {font-family:\"Palatino\";font-size:40pt;text-align:left;}  "))
css_parsing._parse_rule(la,sp)
assert unicoding3_0.equals(sp._font_name,unicoding3_0.string_of("Palatino"))
assert sp._font_size == 40.0
assert len(sp._text_alignment_map) == 1
assert sp._text_alignment_map["p.begin"] == "left"
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("p.end{text-align:right;}"))
css_parsing._parse_rule(la,sp)
assert unicoding3_0.equals(sp._font_name,unicoding3_0.string_of("Palatino"))
assert sp._font_size == 40.0
assert len(sp._text_alignment_map) == 2
assert sp._text_alignment_map["p.begin"] == "left"
assert sp._text_alignment_map["p.end"] == "right"
print("OK")

print("Tests of parse_style_sheet, font_name_of, font_size_of", end=' ')
#sp = css_parsing.parse_style_sheet(None)
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("      "))
sp = css_parsing.parse_style_sheet(la)
#assert unicoding3_0.equals(css_parsing.font_name_of(None),unicoding3_0.string_of("Times New Roman"))
assert unicoding3_0.equals(css_parsing.font_name_of(sp),unicoding3_0.string_of("Times New Roman"))
#assert css_parsing.font_size_of(None) == 12.0
assert css_parsing.font_size_of(sp) == 12.0
assert looking_ahead.current_symbol_of(la) == looking_ahead.END_OF_STREAM
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("  p{font-size:10pt;}  p{font-size:8pt;}  "))
sp = css_parsing.parse_style_sheet(la)
assert unicoding3_0.equals(css_parsing.font_name_of(sp),unicoding3_0.string_of("Times New Roman"))
assert css_parsing.font_size_of(sp) == 8.0
assert looking_ahead.current_symbol_of(la) == looking_ahead.END_OF_STREAM
print("OK")

print("Tests of css_parsing.alignment_of", end=' ')
la = looking_ahead.lookahead_of_string(unicoding3_0.string_of("p.end{text-align:right;}"))
sp = css_parsing.parse_style_sheet(la)
#assert unicoding3_0.equals(css_parsing.alignment_of(None,unicoding3_0.string_of("p.end")),unicoding3_0.string_of("right"))
#assert unicoding3_0.equals(css_parsing.alignment_of(sp,None),unicoding3_0.string_of("right"))
assert unicoding3_0.equals(css_parsing.alignment_of(sp,unicoding3_0.string_of("p.end")),unicoding3_0.string_of("right"))
assert unicoding3_0.equals(css_parsing.alignment_of(sp,unicoding3_0.string_of("p")),unicoding3_0.string_of("left"))
print("OK")

print("")
print("All tests OK")



from enum import Enum
import threading
""" 
Contractor which exposes a type Text, which is a series of lines,
indexed from 0.  Each line is a unicoding3_0.String (mutable sequence of 
unicode code points), indexed from 0.
A text always has a least one line (which may be empty).
Lines are separated by NL ('\n').
Each line has an associated alignment.
A text has an associated cursor, which can be used to access the text sequentially in either direction, or randomly.
The Text can be used as a lookahead, by using the procedures
"current_code_point" and "advance".
The type Text has a lock _my_lock, which should be used in 
multi-threading situations.
"""

# author R.N.Bosworth

# version 27 Feb 2023  15:%8

from guibits1_0 import type_checking2_0, unicoding3_0

"""
Copyright (C) 2012,2015,2016,2017,2018,2019,2021,2022,2023  R.N.Bosworth

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License (gpl.txt) for more details.
"""

# exposed types
# -------------

Alignment = Enum('Alignment','BEGIN MIDDLE END')

class Text:
  """ 
  Invariants:
    the text always has at least one line (which may be empty)
    the cursor always is in a valid position (in the text or at end of text)
  """
  
  def __init__(self):
    """
    pre:
      self = Text to be initialized
      
    post:
      _my_text = list of Lines containing one empty Line
      _cursor_line = 0
      _cursor_char = 0
      _modified = False
      _first_paragraph = True
      
    test:
      twice thru
    """
    self._my_lock = threading.Lock()
    self._my_text = []
    self._my_text.append(_Line())
    self._cursor_line = 0
    self._cursor_char = 0
    self._modified = False
    self._first_paragraph = True


# exposed constants
# -----------------

END_OF_TEXT =  -1
START_OF_TEXT =  -2


# exposed procedures
# ------------------

def advance(t):
  """ 
  pre:
    t = Text whose cursor is to be advanced
    
  post:
     if t's cursor was before the end of Text,
       it has been moved one position towards the end of the text

  test:
    t = None
    t = ""
    t = "a\nb"
      cursor at start of text
        call advance four times
  """
  type_checking2_0.check_derivative(t,Text)  
  line = t._my_text[t._cursor_line]
  if t._cursor_char < unicoding3_0.length_of(line._my_line):
    t._cursor_char += 1
  else:
    if t._cursor_line < (len(t._my_text)-1):
      t._cursor_char = 0
      t._cursor_line += 1
    else:
      pass  # cursor at end of text
    

def current_code_point(t):
  """ 
  pre:  t = Text to be read
        
  post: returns UTF-32 code point of character at cursor position of t,
          or END_OF_TEXT if the cursor is at the end of t, as an int
  test:
    t = None
    t = ""
    t = "a\nb"
      call current_code_point
      call current_code_point,advance four times
  """
  type_checking2_0.check_derivative(t,Text)
  line = t._my_text[t._cursor_line]
  if t._cursor_char < unicoding3_0.length_of(line._my_line):
    return unicoding3_0.code_point_at(line._my_line,t._cursor_char)
  else:
    if t._cursor_line < len(t._my_text) - 1:
      return ord('\n')
    else:
      return END_OF_TEXT
      

def cursor_code_point_offset(t):
  """ 
  pre:
    t = Text whose cursor's char-offset is to be queried
        
  post: 
    char-offset within the line of t's cursor has been returned, as an int
    
  test:
    t = None
    t = valid Text
      cursor char offset 0
      cursor char offset non-zero
  """
  type_checking2_0.check_derivative(t,Text)
  return t._cursor_char
  

def cursor_line_offset(t):
  """ 
  pre:  
    t = Text whose cursor's line-offset is to be queried
        
  post: 
    line-offset of t's cursor has been returned, as an int

  test:
    t = None
    t = valid Text
      zero line length
      non-zero line length
  """
  type_checking2_0.check_derivative(t,Text)
  return t._cursor_line
  

def delete_after(t):
  """ 
  pre:
    t = Text to be modified
    
  post:
    character after cursor of text t has been deleted, 
      if possible
    returns True iff char has been deleted, else False

  post:
    t._modified = True, if this procedure's action modified the text

  test:
    t = None
    t = valid Text
      cursor two chars before end of non-last line
      cursor at end of non-last line
      cursor at end of text
  """
  type_checking2_0.check_derivative(t,Text)
  line = t._my_text[t._cursor_line]
  
  # if at end of text, do nothing
  if t._cursor_line == len(t._my_text) - 1 \
  and t._cursor_char == unicoding3_0.length_of(line._my_line):
    return False
    
  # if at end of line, combine this line with next
  if t._cursor_char == unicoding3_0.length_of(line._my_line):
    line2 = t._my_text[t._cursor_line + 1]
    unicoding3_0.append_a_copy(line._my_line,line2._my_line)
    t._my_text.pop(t._cursor_line + 1)
    t._modified = True
    
  # otherwise, remove the character after the cursor 
  #   and pull the rest of the line down
  else:
    unicoding3_0.remove(line._my_line,t._cursor_char)
    t._modified = True
    
  return True
  

def get_alignment(t):
  """ 
  pre:
    t = Text whose alignment is to be found
    
  post:
    Alignment of line of t containing the cursor has been returned

  test:
    t = None
    t = valid Text
      current line's alignment = BEGIN
      current line's alignment = BEGIN
      current line's alignment = END
  """
  type_checking2_0.check_derivative(t,Text)
  return t._my_text[t._cursor_line]._my_alignment
  

def has_been_modified(t):
  """ 
  pre:
    t = Text to be tested for modification
    
  post:
    returns True iff t has been modified, else False
    
  test:
    t = None
    t = valid Text
      t has been modified
      t has not been modified
  """
  type_checking2_0.check_derivative(t,Text)
  return t._modified
  

def insert_code_point(t,cp):
  """ 
  pre:  t = Text to be modified
        cp = code point of unicode character to be inserted, as an int
        
  post: character with code point cp has been inserted 
          at cursor position of text t
        cursor position has been incremented

  post:
    t._modified = true

  test:
    insert_code_point(None,None)
    t = empty text
      insert_code_point(t,None)
      insert_code_point(t,-1)
      insert_code_point(t,0x110000)
      insert '\n'
      insert 'a'
    t = text of two lines with MIDDLE alignment
      insert '\n' at (0,0)
      insert '\n' at end of first line
      insert '\n' at middle of first line
      insert 'a' at (0,0)
      insert 'a' at end of first line
      insert 'a' at middle of first line
  """
  type_checking2_0.check_derivative(t,Text)
  type_checking2_0.check_identical(cp,int)
  if cp < 0 or cp > 0x10ffff:
    raise Exception("Attempt to write invalid code point: " + hex(cp))
  if cp == ord('\n'):
    # find alignment of current line
    a = t._my_text[t._cursor_line]._my_alignment
    # split line at cursor point
    s = unicoding3_0.python_string_of(t._my_text[t._cursor_line]._my_line)
    s1 = s[0:cursor_code_point_offset(t)]
    s2 = s[cursor_code_point_offset(t):_line_length(t)]
    # old line is replaced by first substring
    line1 = _Line()
    line1._my_alignment = a
    line1._my_line = unicoding3_0.string_of(s1)
    t._my_text[t._cursor_line] = line1
    # insert new line made from second substring
    line2 = _Line()
    line2._my_alignment = a
    line2._my_line = unicoding3_0.string_of(s2)
    t._cursor_line += 1
    t._my_text.insert(t._cursor_line,line2)
    # cursor is set to start of new line
    t._cursor_char = 0
    
  # non-LF char
  else:
    # insert code point
    unicoding3_0.insert(t._my_text[t._cursor_line]._my_line,t._cursor_char,cp)
    # increment cursor
    t._cursor_char += 1
    
  t._modified = True
  

def new_text():
  """ 
  post: a new variable of type Text has been returned.

  test: once thru
  """
  t = Text()
  # text always has at least one line
  return t
  

def retreat(t):
  """ 
  pre:
    t = Text whose cursor is to be retreated
    
  post:
    if t's cursor was before the start of text, 
      it has been moved one position nearer to the start of text

  test:
    t = None
    t = ""
    t = "a\nb"
      cursor at end of text
        call retreat 4 times
  """
  type_checking2_0.check_derivative(t,Text)
  if t._cursor_char == 0:
    if t._cursor_line == 0:
      pass  # cursor at beginning of text
    else:
      t._cursor_line -= 1
      t._cursor_char = _line_length(t)
  else:
    t._cursor_char -= 1
    
  
def set_alignment(t,a):
  """ 
  pre:
    t = Text whose alignment is to be set
    a = Alignment value to which current line of t is to be set
    
  post:
    alignment of line of t containing the cursor has been set to a

  test:
    t = None
    t = valid text
      a = None
    t = valid Text of several lines 
      a = MIDDLE (check alignment)
  """
  type_checking2_0.check_derivative(t,Text)
  type_checking2_0.check_derivative(a,Alignment)
  t._my_text[t._cursor_line]._my_alignment = a
  

def set_cursor(t,lo,co):
  """ 
  pre:  t = Text whose cursor is to be set
        lo = line offset to which cursor is to be set, as an int
        co = character offset to which cursor is to be set, as an int
        
  post: cursor has been set (as far as is possible) to the client's requirements

  test:
    t = None
    t = valid Text of several lines
      lo = None
      lo = -1
      lo = 0
        co = None
        co = -1
        co = 0  (check values)
      lo = one line past end of text
      lo = last line of text
        co = one character past end of text
        co = last character of text  (check values)
  """
  type_checking2_0.check_derivative(t,Text)
  type_checking2_0.check_identical(lo,int)
  if lo < 0 or lo >= len(t._my_text):
    raise Exception("Attempt to set cursor line offset to invalid value: lo="+str(lo))
  type_checking2_0.check_identical(co,int)
  if co < 0 or co > unicoding3_0.length_of(t._my_text[lo]._my_line):
    raise Exception("Attempt to set cursor character offset to invalid value: co="+str(co))
  t._cursor_line = lo
  t._cursor_char = co
  

def set_cursor_end(t):
  """ 
  pre:
    t = Text whose cursor is to be set
        
  post:
    cursor position has been set to the end of the text

  test:
    t = None
    empty text
    non-empty text with non-empty last line (check value)
  """
  type_checking2_0.check_derivative(t,Text)
  t._cursor_line = len(t._my_text) - 1
  line = t._my_text[t._cursor_line]
  t._cursor_char = unicoding3_0.length_of(line._my_line)
  

def set_cursor_start(t):
  """ 
  pre:
    t = Text whose cursor is to be set
        
  post:
    cursor position has been set to the start of the text

  test:
    t = None
    t = valid text (check value)
  """
  type_checking2_0.check_derivative(t,Text)
  set_cursor(t,0,0)
  

def set_modified(t):
  """ 
  pre:
    t = Text to be marked as modified
    
  post:
    t has been marked as modified

  test:
    t = None
    t = valid text (check value)
  """
  type_checking2_0.check_derivative(t,Text)
  t._modified = True
  

def set_unmodified(t):
  """ 
  pre:
    t = Text to be marked as unmodified
    
  post:
    t has been marked as unmodified

  test:
    t = None
    t = valid text (check value)
  """
  type_checking2_0.check_derivative(t,Text)
  t._modified = False
  

# private types
# -------------

class _Line:
  def __init__(self):
    """
    pre:
      self = _Line to be initialized
      
    post:
      _my_aligment = Alignment.BEGIN
      _my_line = unicoding3_0.new_string()
      
    test:
      twice thru
    """
    self._my_alignment = Alignment.BEGIN
    # default alignment for new line
    self._my_line = unicoding3_0.new_string()


# private procedures
# ------------------

def _line_length(t):
  """ 
  pre:
    t = text whose current line length is to be returned
    
  post:
    returns length of current line of t (line where cursor is positioned)
    Note: the length does not include the final '\n', if present

  test:
    once thru
  """
  return unicoding3_0.length_of(t._my_text[t._cursor_line]._my_line)



from guibits1_0 import type_checking2_0
from . import code_point_testing
from . import html_css_parsing
from . import looking_ahead
from guibits1_0 import unicoding3_0

# author R.N.Bosworth

# version 9 Mar 23  15:45
""" 
Contractor for parsing Cascading Style Sheets (simplified syntax!).

Copyright (C) 2016,2017,2018,2021,2022,2023  R.N.Bosworth

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License (gpl.txt) for more details.

  CSS Grammar version 17.3.18   15:54

  This grammar is a reduced version of the W3C CSS standard, as at April 2016.

  This grammar is in EBNF (Extended Backus-Naur Form) using the following meta-symbols:  ::=  is  (  )  |  *.  The equivalent terminal characters are called LPARENTHESIS, RPARENTHESIS, BAR and STAR.

  Non-terminals are in CamelCase.  Terminal names are in UPPER_CASE.  Other symbols represent themselves.
  
  The rules are case-sensitive, unlike CSS.
    

  StyleSheet  ::=  OptionalSpaces Rule*
  
  Rule  ::=  Selector OptionalSpaces DeclarationBlock OptionalSpaces
  
  Selector  :==  DotName
  
  DotName  ::=  Name ( . Name | )
                   
  DeclarationBlock  ::=  { OptionalSpaces Declaration OptionalSpaces ; OptionalSpaces 
                          (Declaration OptionalSpaces ; OptionalSpaces)* }

  Declaration  ::=  FontDeclaration | TextAlignmentDeclaration
  
  FontDeclaration  ::=  f o n t - (FontNameDeclarationRest | FontSizeDeclarationRest)
    
  FontNameDeclarationRest  ::=  f a m i l y OptionalSpaces : OptionalSpaces FontName
  
  FontName  ::=  " FontNameCharacter FontNameCharacter* "
  
  FontNameCharacter  ::=   A .. Z | a .. z | SPACE
  
  FontSizeDeclarationRest  ::=  s i z e OptionalSpaces : OptionalSpaces Digit ( | Digit) p t
  
  Digit  ::=  0..9
  
  TextAlignmentDeclaration  ::=  t e x t - a l i g n OptionalSpaces : OptionalSpaces TextAlignValue
  
  TextAlignValue  ::=  l e f t | r i g h t | c e n t e r | j u s t i f y 

  (Note that "left" and "right" in this context means "the beginning of the line" and "the end of the line", the actual positioning depending on whether the text is read left-to-right or right-to-left.)
  
  Rules for OptionalSpaces and Name are defined in HTMLCSSGrammar.
"""

# exposed types
# -------------

class StyleProperties:
  pass  

# exposed procedures
# ------------------

def alignment_of(sp,s):
  """ 
  pre:
    sp = StyleProperties variable whose alignment properties are to be accessed
    s = unicoding3_0.String whose alignment in sp is to be found
  
  post:
    CSS alignment value corresponding to s in sp has been returned,
      or "left" if no match for s is found

  test:
    sp with mapping from "p.end" to "right"
      s = "p.end"
      s = "p"
  """
  type_checking2_0.check_derivative(sp,StyleProperties)
  type_checking2_0.check_derivative(s,unicoding3_0.String)
  js = sp._text_alignment_map.get(unicoding3_0.python_string_of(s))
  if js == None:
    return unicoding3_0.string_of("left")
  else:
    return unicoding3_0.string_of(js)
    
  
def font_name_of(sp):
  """ 
  pre:
    sp = StyleProperties variable to be examined
  
  post:
    font name of sp has been returned

  test:
    once thru
  """
  type_checking2_0.check_derivative(sp,StyleProperties)  
  return sp._font_name
  

def font_size_of(sp):
  """ 
  pre:
    sp = StyleProperties variable to be examined
  
  post:
    font size of sp has been returned

  test:
    once thru
  """
  type_checking2_0.check_derivative(sp,StyleProperties)
  return sp._font_size
  

def parse_style_sheet(la):
  """ 
  pre:
    la = lookahead which is to be parsed
  
  post:
    returns StyleProperties variable corresponding to StyleSheet which has been parsed
      (empty StyleSheet gives default StyleProperties variable)
    la has been advanced past StyleSheet

  syntax:
    StyleSheet  ::=  OptionalSpaces Rule*

  test:
    "      "
    "  p{font-size:10pt;}  p{font-size:8pt;}  "
  """
  type_checking2_0.check_derivative(la,looking_ahead.Lookahead)
  sp = _new_style_properties()
  sp._font_name = unicoding3_0.string_of("Times New Roman")
  sp._font_size = 12.0
  html_css_parsing.accept_optional_spaces(la)
  while code_point_testing.is_name_char(looking_ahead.current_symbol_of(la)):
    _parse_rule(la,sp)
  return sp
  

# private members
# ---------------

def _is_digit(cp):
  """ 
  pre:
    cp = code point to be tested
  
  post:
    returns true iff cp is a Digit

  syntax:
    Digit  ::=  0..9
                ( U+0030 .. U+0039 )

  test:
    cp = '/'
         '0'
         '9'
         ':'
  """
  return (ord('0') <= cp and cp <= ord('9'))
  

def _is_font_name_char(cp):
  """ 
  pre:
    cp = code point to be tested
  
  post:
    returns true iff cp is a CSS font name character

  syntax:
    FontNameCharacter  ::=   A .. Z | a .. z | SPACE

  test:
    cp = @
         A
         Z
         [
         U+0060
         a
         z
         {
         U+001F
         U+0020
         U+0021
  """
  return (ord('A') <= cp and cp <= ord('Z') or ord('a') <= cp and cp <= ord('z') or cp == ord(' '))
  

def _new_style_properties():
  """
  pre:
    True
    
  post:
    a new StyleProperties variable has been returned,
      with default values
      
  test:
    once thru (check values)
  """
  sp = StyleProperties()
  sp._text_alignment_map = dict()  
  # map between selector names and alignment values:
  #   left, right, center, justify
  sp._font_name = None
  sp._font_size = 0.0
  return sp


def _parse_declaration(la,sp,s):
  """ 
  pre:
    la = lookahead which is being parsed
    sp = StyleProperties variable to be updated
    s = selector name for this Declaration
  
  post:
    either
      sp has been updated with contents of Declaration
      lookahead has been advanced past Declaration
    or:
      RuntimeException has been thrown
 
  syntax:
    Declaration  ::=  FontDeclaration | TextAlignmentDeclaration

  test:
    s = "body"
      la = "font-family:\"Times New Roman\""
    s = "p.begin"
      la = "text-align:left"
    s = "p.end"
      la = "complete-load-of-rubbish"
  """
  if looking_ahead.current_symbol_of(la) == ord('f'):
    _parse_font_declaration(la,sp)
  elif looking_ahead.current_symbol_of(la) == ord('t'):
    _parse_text_alignment_declaration(la,s,sp._text_alignment_map)
  else:
    raise Exception("Unexpected declaration: expected FontDeclaration or TextAlignmentDeclaration, " + "found declaration beginning with U+" + format(looking_ahead.current_symbol_of(la),'x'))
  

def _parse_declaration_block(la,sp,s):
  """ 
  pre:
    la = lookahead which is being parsed
    sp = StyleProperties variable to be updated
    s = selector name for this DeclarationBlock

  post:
    either
      sp has been updated with contents of DeclarationBlock
      lookahead has been advanced past DeclarationBlock
    or:
      RuntimeException has been thrown

  syntax:
    DeclarationBlock  ::=  { OptionalSpaces Declaration OptionalSpaces ; OptionalSpaces 
                            (Declaration OptionalSpaces ; OptionalSpaces)* }

  test:
    [  font-family:"Times New Roman"}
    {  zont-family:"Times New Roman"}
    {  font-family:"Times New Roman"  ]
    {  font-family:"Times New Roman"  }
    {font-family:"Times New Roman";}
    {  font-family:"Times New Roman"  ; font-family:"Palatino"  }
    {  font-family:"Times New Roman"  ; font-family:"Palatino"  ;  }
    {  
      font-family:"Times New Roman"  ; 
      font-size:10pt  ;  
      font-family:"Palatino"  ; 
      font-size:20pt    ;
      text-align:left;
      text-align:center;       
    }
  """
  html_css_parsing.accept(ord('{'),la)
  html_css_parsing.accept_optional_spaces(la)
  _parse_declaration(la,sp,s)
  html_css_parsing.accept_optional_spaces(la)
  html_css_parsing.accept(ord(';'),la)
  html_css_parsing.accept_optional_spaces(la)
  while looking_ahead.current_symbol_of(la) != ord('}'):
    _parse_declaration(la,sp,s)
    html_css_parsing.accept_optional_spaces(la)
    html_css_parsing.accept(ord(';'),la)
    html_css_parsing.accept_optional_spaces(la)
  html_css_parsing.accept(ord('}'),la)
  

def _parse_dot_name(la):
  """ 
  pre:
    la = lookahead which is being parsed
  
  post:
    either:
      returns HTML dot-name as lower-case unicoding3_0.String
      lookahead has been advanced past dot-name
    or:
      RuntimeException has been thrown

  syntax:
    DotName  ::=  Name ( . Name | )

  test:
    "%"
    "f"
    "fred%"
    "fred.jim"
  """
  s = html_css_parsing.parse_name(la)
  if looking_ahead.current_symbol_of(la) == ord('.'):
    unicoding3_0.append(s,(ord('.')))
    looking_ahead.advance(la)
    unicoding3_0.append_a_copy(s,html_css_parsing.parse_name(la))
  return s
  

def _parse_font_declaration(la,sp):
  """ 
  pre:
    la = lookahead which is being parsed
    sp = StyleProperties variable to be updated
  
  post:
    either:
      sp has been updated with contents of FontDeclaration
      lookahead has been advanced past FontDeclaration
    or:
      RuntimeException has been thrown

  syntax:
    FontDeclaration  ::=  f o n t - (FontNameDeclarationRest | FontSizeDeclarationRest)

  test:
    fonu-family : "Times New Roman"
    font=family : "Times New Roman"
    font-family : "Times New Roman"
    font-size : 42.0;
    font-blob : "blobbie"
  """
  html_css_parsing.accept_name(unicoding3_0.string_of("font"),la)
  html_css_parsing.accept(ord('-'),la)
  if looking_ahead.current_symbol_of(la) == ord('f'):
    sp._font_name = _parse_font_name_declaration_rest(la)
  elif looking_ahead.current_symbol_of(la) == ord('s'):
    sp._font_size = _parse_font_size_declaration_rest(la)
  else:
    raise Exception("unexpected code point, expected U+" + format(ord('f'),'x') + " or U+" + format(ord('s'),'x') + ", found U+" + format(looking_ahead.current_symbol_of(la),'x'))
  

def _parse_font_name(la):
  """ 
  pre:
    la = lookahead which is being parsed
  
  post:
    either:
      returns font name as unicoding3_0.String
      lookahead has been advanced past FontName
    or:
      RuntimeException has been thrown

  syntax:
    FontName  ::=  " FontNameCharacter FontNameCharacter* "

  test:
    ''
    ""
    "a"
    "a%"
    "Times Roman'
    "Times Roman"
  """
  s = unicoding3_0.new_string()
  html_css_parsing.accept(ord('\"'),la)
  if  not _is_font_name_char(looking_ahead.current_symbol_of(la)):
    raise Exception("unexpected code point in font name: U+" + format(looking_ahead.current_symbol_of(la),'x'))
  unicoding3_0.append(s,(looking_ahead.current_symbol_of(la)))
  looking_ahead.advance(la)
  while _is_font_name_char(looking_ahead.current_symbol_of(la)):
    unicoding3_0.append(s,(looking_ahead.current_symbol_of(la)))
    looking_ahead.advance(la)
  html_css_parsing.accept(ord('\"'),la)
  return s
  

def _parse_font_name_declaration_rest(la):
  """ 
  pre:
    la = lookahead which is being parsed
  
  post:
    either:
      returns font name as unicoding3_0.String
      lookahead has been advanced past FontNameDeclaration
    or:
      RuntimeException has been thrown

  syntax:
    FontNameDeclarationRest  ::=  f a m i l y OptionalSpaces : OptionalSpaces FontName

  test:
    familz:"New Century Schoolbook"
    family;"New Century Schoolbook"
    family:"New Century Schoolbook"
    family  :  "Gill Sans"
  """
  html_css_parsing.accept_name(unicoding3_0.string_of("family"),la)
  html_css_parsing.accept_optional_spaces(la)
  html_css_parsing.accept(ord(':'),la)
  html_css_parsing.accept_optional_spaces(la)
  return _parse_font_name(la)
  

def _parse_font_size_declaration_rest(la):
  """ 
  pre:
    la = lookahead which is being parsed
  
  post:
    either:
      returns font size in points as double
      lookahead has been advanced past FontSizeDeclaration
    or:
      RuntimeException has been thrown

  syntax:
    FontSizeDeclarationRest  ::=  s i z e OptionalSpaces : OptionalSpaces Digit ( | Digit) p t

  test:
    tize:0pt
    sjze:0pt
    siye:0pt
    sizf:0pt
    size;0pt
    size:/pt
    size:0pt
    size:9pt
    size::pt
    size:1/pt
    size:10pt
    size:19pt
    size:1:pt
    size:99qt
    size:99pu
    size:99pt
    size  :  99pt
  """
  html_css_parsing.accept_name(unicoding3_0.string_of("size"),la)
  html_css_parsing.accept_optional_spaces(la)
  html_css_parsing.accept(ord(':'),la)
  html_css_parsing.accept_optional_spaces(la)
  if  not _is_digit(looking_ahead.current_symbol_of(la)):
    raise Exception("unexpected code point in number: U+" + format(looking_ahead.current_symbol_of(la),'x'))
  d = looking_ahead.current_symbol_of(la) - ord('0')
  looking_ahead.advance(la)
  if _is_digit(looking_ahead.current_symbol_of(la)):
    d = d*10.0 + (looking_ahead.current_symbol_of(la) - ord('0'))
    looking_ahead.advance(la)
  html_css_parsing.accept(ord('p'),la)
  html_css_parsing.accept(ord('t'),la)
  return d
  

def _parse_rule(la,sp):
  """ 
  pre:
    la = lookahead which is being parsed
    sp = StyleProperties variable to be updated
  
  post:
    either
      sp has been updated with contents of Rule
      lookahead has been advanced past Rule
    or:
      RuntimeException has been thrown

  syntax:
    Rule  ::=  Selector OptionalSpaces DeclarationBlock OptionalSpaces
  
    Selector  :==  DotName

  test:
    "p.begin   {font-family:\"Palatino\";font-size:40;text-align:left;}  "
    "p.end{text-align:right;}"
  """
  s = _parse_dot_name(la)
  html_css_parsing.accept_optional_spaces(la)
  _parse_declaration_block(la,sp,s)
  html_css_parsing.accept_optional_spaces(la)
  

def _parse_text_alignment_declaration(la,s,m):
  """ 
  pre:
    la = lookahead which is being parsed
    s = selector name for the current rule being parsed
    m = map which contains the current mappings between selector names and alignments
  
  post:
    m has been updated with a new mapping from s to the parsed alignment value
    la has been advanced past the text alignment declaration

  syntax:
    TextAlignmentDeclaration  ::=  t e x t - a l i g n Spaces : Spaces TextAlignValue

  test:
    s = "fred"
      la = "text-aligm:left"
           "text-align:left"
    s = "jim"
      la = "text-align   :   justify"
  """
  html_css_parsing.accept_name(unicoding3_0.string_of("text"),la)
  html_css_parsing.accept(ord('-'),la)
  html_css_parsing.accept_name(unicoding3_0.string_of("align"),la)
  html_css_parsing.accept_optional_spaces(la)
  html_css_parsing.accept(ord(':'),la)
  html_css_parsing.accept_optional_spaces(la)
  av = _parse_text_align_value(la)
  m[unicoding3_0.python_string_of(s)] = unicoding3_0.python_string_of(av)
  

def _parse_text_align_value(la):
  """ 
  pre:
    la = lookahead which is being parsed
    sp = StyleProperties variable to be updated
  
  post:
    either:
      returns text align value as lower case unicoding3_0.String
      lookahead has been advanced past text align value
    or:
      RuntimeException has been thrown

  syntax:
    TextAlignValue  ::=  l e f t | r i g h t | c e n t e r | j u s t i f y

  test:
    "bollocks"
    "left"
    "LEFT"
    "RIGHT"
    "right"
    "center"
    "CENTER"
    "JUSTify"
    "just_i_f_y"
  """
  s = html_css_parsing.parse_name(la)
  if unicoding3_0.equals(s,unicoding3_0.string_of("left")) or \
     unicoding3_0.equals(s,unicoding3_0.string_of("right")) or \
     unicoding3_0.equals(s,unicoding3_0.string_of("center")) or \
     unicoding3_0.equals(s,unicoding3_0.string_of("justify")):
    return s
  else:
    raise Exception("Unexpected alignment value: " + unicoding3_0.python_string_of(s))



from . import texting
from guibits1_0 import type_checking2_0

# author R.N.Bosworth

# version 2 Mar 2023  15:15

""" 
Contractor which allows the client to advance and retreat through HTML text, 
as it appears on the screen,
ignoring tags and dealing with escaped code points.

An "HTML character" is either an escaped or non-escaped Unicode code point.

HTML tags are ignored by this contractor.

So

  "<b>1 &lt; 2</b>" is treated as "1 < 2"

Copyright (C) 2018,2021  R.N.Bosworth

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License (gpl.txt) for more details.


"""
""" 
  StartTag(Name)  ::=  < Name OptionalSpaces >

  EndTag(Name)  ::=  < / Name OptionalSpaces >

  EscapedHtmlCharacter  ::=  ( & a m p ;
                             | & g t ;
                             | & l t ;
                             | any Unicode code point except '&' (U+26), '>' (U+3E) and '<' (U+3C)
"""

# exposed procedures
# ------------------

def advance(t):
  """ 
  pre:
    t = texting.Text whose cursor is to be advanced
    
  post:
    if t is at end of text, false has been returned
    otherwise,
      true has been returned
      t's cursor has been moved one position towards the end of the text,
        ignoring any HTML tags and treating escaped code points as one position

  test:
    t = None
    t = "<b><i>1&lt;2</i></b>x<b></b>"
      cursor at (0,0)
        call advance as many times as necessary!
  """
  type_checking2_0.check_derivative(t,texting.Text)
  line_offset = texting.cursor_line_offset(t)
  code_point_offset = texting.cursor_code_point_offset(t)
  texting.set_cursor_end(t)
  if line_offset == texting.cursor_line_offset(t) and code_point_offset == texting.cursor_code_point_offset(t):
    # at end of text
    return False
    
  texting.set_cursor(t,line_offset,code_point_offset)
  cp = texting.current_code_point(t)
  texting.advance(t)
  if cp == ord('<'):
    while texting.current_code_point(t) != ord('>'):
      texting.advance(t)
      
    texting.advance(t)
    # text cursor just past '>'
    return advance(t)
    
  if cp == ord('&'):
    while texting.current_code_point(t) != ord(';'):
      texting.advance(t)
      
    texting.advance(t)
    # text cursor just past ';'
    return True
    
  # else cp is a LF, a SPACE or a NonSpaceChar
  return True
  

def delete_after(t):
  """ 
  pre:
    t = texting.Text for which the HTML character following the cursor is to be deleted
    
  post:
    if t is at end of text, false has been returned
    otherwise,
      true has been returned
      the HTML character following t's cursor has been deleted

  test:
    t = None
    t = "<b><i>1&lt;2</i></b>x<b></b>"
      cursor at (0,0)
        call delete_after as many times as necessary!
  """
  type_checking2_0.check_derivative(t,texting.Text)
  line_offset = texting.cursor_line_offset(t)
  code_point_offset = texting.cursor_code_point_offset(t)
  texting.set_cursor_end(t)
  if line_offset == texting.cursor_line_offset(t) and code_point_offset == texting.cursor_code_point_offset(t):
    # at end of text
    return False
    
  texting.set_cursor(t,line_offset,code_point_offset)
  if texting.current_code_point(t) == ord('<'):
    texting.advance(t)
    while texting.current_code_point(t) != ord('>'):
      texting.advance(t)
      
    texting.advance(t)
    # text cursor just past '>'
    return delete_after(t)
    
  if texting.current_code_point(t) == ord('&'):
    texting.delete_after(t)
    while texting.current_code_point(t) != ord(';'):
      texting.delete_after(t)
      
    texting.delete_after(t)
    # escaped HTML character has been deleted
    return True
    
  # else cp is a LF, a SPACE or a NonSpaceChar
  texting.delete_after(t)
  return True
  

def retreat(t):
  """ 
  pre:
    t = texting.Text whose cursor is to be retreated
    
  post:
    if t is at start of text, false has been returned
    otherwise,
      true has been returned
      t's cursor has been moved one position towards the start of the text,
        ignoring any HTML tags and treating escaped code points as one position

  test:
    t = None
    t = "x;<b><i>1&lt;; ;2</i></b>y"
      cursor at end of text
        call retreat as many times as necessary!
    t = "<align>x"
      cursor at end of text
        call retreat as many times as necessary!
  """
  type_checking2_0.check_derivative(t,texting.Text)
  if texting.cursor_line_offset(t) == 0 and texting.cursor_code_point_offset(t) == 0:
    return False
    
  texting.retreat(t)
  if texting.current_code_point(t) == ord('>'):
    texting.retreat(t)
    while texting.current_code_point(t) != ord('<'):
      texting.retreat(t)
      
    # text cursor just before '<'
    return retreat(t)
    
  if texting.current_code_point(t) == ord(';'):
    semicolon_position = texting.cursor_code_point_offset(t)
    # in case it really is a semicolon
    texting.retreat(t)
    cp = texting.current_code_point(t)
    while cp != ord('&') and cp != ord(';') and cp != ord(' ') and cp != ord('>') and texting.cursor_code_point_offset(t) > 0:
      # to ensure termination
      texting.retreat(t)
      cp = texting.current_code_point(t)
      
    # text cursor EITHER just before '&' OR just before ';' OR just before ' ' OR
    # just before '>' OR at start of line 
    # (note ';', ' ' and '>' not allowed in escaped code point)
    if cp != ord('&'):
      texting.set_cursor(t,texting.cursor_line_offset(t),semicolon_position)
      # just before actual ';', as no '&' found
      
    # just before previous code point, escaped or not
    return True
    
  # else current_code_point is a non-escaped HTML Character
  return True
